from __future__ import annotations
from rich import print
import argparse


class AnalyzerConfig:
    gauss_size: tuple[int,int] = (5,5)
    "gaussian filter size before adaptive thold"
    gauss_sigma: float = 5.0
    "gaussian sigma before adaptive thold"

    thresh_block_size: int = 11
    "adaptive threshold block size"
    thresh_sensitivity: float = 5.0
    "adaptive threshold sensitivity"

    skelclose_size: int = 3
    "size of final closing kernel for skeleton"
    skelclose_amnt: int = 5
    "iteration count of final closing kernel for skel"

    fragment_min_area_px: int = 20
    "minimum fragment area"
    fragment_max_area_px: int = 25000
    "maximum fragment area"

    size_factor: float = 1.0
    "factor to scale fragment size in mm/px"
    real_image_size: tuple[int,int] = None
    "real image size in mm"
    cropped_image_size: tuple[int,int] = None
    "real image size in mm"
    crop: bool = False
    "crop input image"
    impact_position: tuple[float, float] = None
    "impact position in mm [X Y]"


    debug: bool = False
    "enable debug output"
    debug_experimental: bool = False
    "enable debug output"
    display_region: tuple[int,int,int,int]  = None
    "region to display in output plots (x1,y1,x2,y2)"
    resize_factor: float = 1.0
    "factor to resize input image in preprocess"
    displayplots: bool = False
    "Display plots during creation"
    printconfig: bool = False
    "Print the configuration before starting the script"

    out_name: str = ""
    "name of the output directory"
    ext_plots: str = "png"
    "output extension for plots"
    ext_imgs: str = "png"
    "output extension for images"

    skip_darkspot_removal: bool  = False
    "skip dark spot removal"
    intensity_h: int = 500
    "intensity kernel width in px"

    path: str = ""
    "Path to data"

    norm_region_center: tuple[float, float] = None
    "Center for evaluation region according to DIN in mm [X Y]"

    norm_region_size: tuple[float, float] = (50,50)
    "Size for evaluation region according to DIN in mm [W H]"



    def get_parser(descr) -> argparse.ArgumentParser:
        """
        Create and return an argumentParser, that can be used to initialize a new
        AnalyzerConfig with `AnalyzerConfig.from_args(args)` method.

        This can be used, if the argumentparser should be extended. I.e. if highspeed
        module wants to add a specific argument group.

        Args:
            descr (string): The description, the argparse object should display on `-h`.

        Returns:
            argparse.ArgumentParser: Can be used to create AnalyzerConfig.
        """
        parser = argparse.ArgumentParser(description=descr, formatter_class=argparse.RawDescriptionHelpFormatter)

        gnrl_group = parser.add_argument_group("General")
        gnrl_group.add_argument('--displayplots', action='store_true', \
            help='Instruct the analyzer to display output plots.', default=False)
        gnrl_group.add_argument('--debug', action='store_true', \
            help='Sets a debug flag to display verbose output.', default=False)
        gnrl_group.add_argument('--exp-debug', action='store_true', \
            help='Sets an experimental debug flag to display verbose output.', default=False)
        gnrl_group.add_argument('--printconfig', action='store_true', \
            help='Print the config before starting the script.', default=False)
        gnrl_group.add_argument('-display-region', nargs=4, help='Region to display in debug outputs. [Pixels]',\
            type=int, default=None, metavar=('X1', 'Y1', 'X2', 'Y2'))

        imgroup = parser.add_argument_group("Image operations")
        imgroup.add_argument('path', nargs="?", help='The path of the image to be processed or a folder that contains a file in subfolder "[path]/fracture/morph/...Transmission.bmp".')
        imgroup.add_argument('-realsize', nargs="*", help='Real size of the input image. If only one dim is provided, a square geometry is used.',\
            type=int, default=None, metavar=('WIDTH', 'HEIGHT'))
        imgroup.add_argument('-cropsize', nargs="*", help='Crop image size in pixels. If only one dim is provided, a square geometry is used.',\
            type=int, default=None, metavar=('WIDTH', 'HEIGHT'))

        prep = parser.add_argument_group("Preprocessor")
        # image preprocessing arguments
        prep.add_argument('-gauss-size', help='Gaussian filter size',\
            type=int, default=5)
        prep.add_argument('-gauss-sigma', help='Gaussian filter sigma',\
            type=float, default=5)
        prep.add_argument('-min-area', help='Minimum fragment area threshold [px²]',\
            type=float, default=20)
        prep.add_argument('-max-area', help='Maximum fragment area threshold [px²]',\
            type=float, default=25000)
        prep.add_argument('-thresh-sens', help='Adaptive threshold sensitivity',\
            type=float, default=6)
        prep.add_argument('-thresh-block', help='Adaptive threshold block size',\
            type=int, default=11)
        prep.add_argument('-resize-fac', help='Image resize factor before adaptive th.',\
            type=float, default=1.0)

        post = parser.add_argument_group("Postprocessor")
        post.add_argument('-skelclose-sz', help='Size for final skeleton close kernel.',\
            type=int, default=3)
        post.add_argument('-skelclose-amnt', help='Iterations for final skeleton close kernel.',\
            type=int, default=5)
        post.add_argument('--skip-spot-elim', help='Instruct the postprocessor to skip "dark-spot" removal.',\
            action="store_true", default=False)
        post.add_argument('-intensity-width', help='Pixel width for intensity calculation.',\
            type=int, default=500)
        post.add_argument('-impactposition', nargs=2, metavar=('X', 'Y'), type=float,
            help='Impact position in mm [X Y]', default=(50,50))
        post.add_argument('-normregioncenter', nargs=2, metavar=('X', 'Y'), type=float, default=None,
            help='Center for evaluation region according to DIN in mm [X Y]')
        post.add_argument('-normregionsize', nargs=2, metavar=('W', 'H'), type=float, default=(50,50),
            help='Size for evaluation region according to DIN in mm.')


        output_group = parser.add_argument_group("Output")
        output_group.add_argument('-out', nargs="?", help='Output directory path.', \
            default="fracsuite-output")
        output_group.add_argument('-plot-ext', nargs="?", help='Plot file extension. Default: png.', \
            default="png", choices=['png', 'pdf', 'jpg', 'bmp'])
        output_group.add_argument('-image-ext', nargs="?", help='Image file extension. Default: png.',\
            default="png", choices=['png', 'jpg', 'bmp'])

        return parser

    def from_args(args: argparse.Namespace) -> AnalyzerConfig:
        """
        Create AnalyzerConfig from command line arguments.

        Args:
            args (argparse.Namespace): The arguments parsed from argparse.

        Returns:
            AnalyzerConfig: The configuration.
        """
        cfg = AnalyzerConfig()
        cfg.debug = args.debug
        cfg.debug_experimental = args.exp_debug
        cfg.printconfig = args.printconfig
        cfg.displayplots = args.displayplots

        cfg.gauss_size = (args.gauss_size,args.gauss_size)
        cfg.gauss_sigma = args.gauss_sigma
        cfg.fragment_min_area_px = args.min_area
        cfg.fragment_max_area_px = args.max_area


        cfg.thresh_block_size = args.thresh_block
        cfg.thresh_sensitivity = args.thresh_sens
        cfg.skelclose_size = args.skelclose_sz
        cfg.skelclose_amnt = args.skelclose_amnt
        cfg.crop = args.cropsize is not None
        cfg.cropped_image_size = args.cropsize
        cfg.real_image_size = args.realsize
        cfg.resize_factor = args.resize_fac
        cfg.impact_position = args.impactposition

        cfg.display_region = args.display_region
        cfg.intensity_h = args.intensity_width
        cfg.ext_plots = args.plot_ext
        cfg.ext_imgs = args.image_ext
        cfg.skip_darkspot_removal = args.skip_spot_elim

        cfg.norm_region_center = args.normregioncenter
        cfg.norm_region_size = args.normregionsize

        if args.debug is True:
            cfg.displayplots = True

        if args.realsize is not None and len(args.realsize) > 1:
            cfg.real_image_size = tuple(args.realsize)
        elif args.realsize is not None and len(args.realsize) == 1:
            cfg.real_image_size = (args.realsize[0], args.realsize[0])

        if args.cropsize is not None and len(args.cropsize) > 1:
            cfg.cropped_image_size = tuple(args.cropsize)
        elif args.cropsize is not None and len(args.cropsize) == 1:
            cfg.cropped_image_size = (args.cropsize[0], args.cropsize[0])

        cfg.path = args.path

        if cfg.printconfig:
            cfg.print()

        return cfg

    def parse(descr: str) -> tuple[argparse.Namespace, AnalyzerConfig]:
        """Directly parse and return a Namespace and AnalyzerConfig.

        Args:
            descr (str): Display description on `-h`.

        Returns:
            tuple[argparse.Namespace, AnalyzerConfig]: Results.
        """
        args = AnalyzerConfig.get_parser(descr).parse_args()
        return args, AnalyzerConfig.from_args(args)

    def __init__(self):
        pass

    def print(self):
        print(self.__dict__)


import typer
import xlsxwriter
import os

app = typer.Typer()


@app.command()
def summarize(path: str):
    """Display a table that summarizes the contents of a database folder.

    Args:
        path (str): Path of the DB.
    """
    workbook = xlsxwriter.Workbook(os.path.join(path,'summary.xlsx'))
 
    # The workbook object is then used to add new
    # worksheet via the add_worksheet() method.
    worksheet = workbook.add_worksheet()
    
    
    worksheet.write(0, 0, "Boundary: A (allseitig), Z (zweiseitig), B (gebettet)")
    worksheet.write(1, 0, "Comment: B (Bohrung)")
    
    start_row = 10
    
    worksheet.write(start_row, 0, "Name")
    worksheet.write(start_row, 1, "Thickness")
    worksheet.write(start_row, 2, "Pre-Stress")
    worksheet.write(start_row, 3, "Boundary")
    worksheet.write(start_row, 4, "Nbr")    
    worksheet.write(start_row, 5, "Comment")
    
    row = start_row + 1
    for folder in os.listdir(path):
        if folder.count('.') < 3:
            continue
        folder_path = os.path.join(path, folder)
        
        # extract data
        sp = folder.split('.')
        thick = sp[0]
        sigma = sp[1]
        bound = sp[2]
        lsp = sp[3].split('-')
        lfnr = lsp[0]
        comment = lsp[1] if len(lsp) > 1 else ""
        
        worksheet.write(row, 0, folder)    
        worksheet.write(row, 1, thick)    
        worksheet.write(row, 2, sigma)    
        worksheet.write(row, 3, bound)    
        worksheet.write(row, 4, lfnr)    
        worksheet.write(row, 5, comment)    
        
        
        
        
        row += 1
    
    # Finally, close the Excel file
    # via the close() method.
    workbook.close()

app()


import os
import shutil
from reportlab.lib.pagesizes import A4
from reportlab.lib.units import mm
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfgen import canvas
from reportlab.pdfbase.ttfonts import TTFont
from tqdm import tqdm
from pylibdmtx.pylibdmtx import encode, ENCODING_SIZE_NAMES, decode
from PIL import Image

__path__ = os.path.dirname(__file__)

# dimensions of A4 paper in mm
A4_WIDTH = 210.0
A4_HEIGHT = 297.0

# dimensions of cell in mm
CELL_WIDTH = 70.0
CELL_HEIGHT = 37.0

# number of cells in each row and column
ROWS = int(A4_HEIGHT/CELL_HEIGHT)  # // is floor division
COLUMNS = int(A4_WIDTH/CELL_WIDTH)

CANVAS_FONT = "F25_Bank_Printer_Bold"
# CANVAS_FONT = "Consolas"
CANVAS_FONT_SIZE = 24

BARCODE_HEIGHT_FAC = 0.6
LABEL_HEIGHT_FAC = 1 - BARCODE_HEIGHT_FAC


def create_datamatrix_code(code: str, label: str) -> str:
    encoded = encode(code.encode('utf-8'), size='RectAuto')    
    img = Image.frombytes('RGB', (encoded.width, encoded.height), encoded.pixels)
    os.makedirs('.out', exist_ok=True)
    name = f'.out/{label}.png'
    img.save(name)
    
    content = decode(img, max_count=1)
    if content[0].data.decode('utf-8') != code:        
        print(f"Decoded code {content[0].data.decode('utf-8')} does not match original code {code}!")
    
    return os.path.abspath(name)
    
def create_cell(label: str, code: str, x: float, y: float, canvas: canvas.Canvas):

    # save barcode image and get its path
    fullname = CREATE_CODE(code, label)

    canvas.setFont(CANVAS_FONT, CANVAS_FONT_SIZE)
    # draw label and barcode on canvas at provided coordinates
    canvas.drawImage(fullname, x*mm , (y + 0.95 * CELL_HEIGHT * LABEL_HEIGHT_FAC)*mm, width = CELL_WIDTH*mm, height = CELL_HEIGHT*mm * BARCODE_HEIGHT_FAC, anchor='n', preserveAspectRatio=ASPECT_RATIO)    
    canvas.drawCentredString((x + CELL_WIDTH / 2)*mm, (y + CELL_HEIGHT * LABEL_HEIGHT_FAC / 2)*mm , label, )
    # canvas.rect(x*mm , y*mm, CELL_WIDTH * mm, CELL_HEIGHT * mm, stroke = 1, fill = 0)
    
    
def generate_pdf(labels_codes: list, filename: str):
    
    fontpath= os.path.join(__path__, f"{CANVAS_FONT}.ttf")
    pdfmetrics.registerFont(TTFont(CANVAS_FONT,fontpath))
    pdfmetrics.registerFontFamily(CANVAS_FONT)

    c = canvas.Canvas(filename, pagesize=(210*mm, 297*mm))

    # coordinates of top left corner of each cell
    coordinates = [(x*CELL_WIDTH, (ROWS-1-y)*CELL_HEIGHT+0.5) for y in range(ROWS) for x in range(COLUMNS)]

    # create all cells
    for i, (label, code) in enumerate(tqdm(labels_codes)):
        # start a new page after every 24 cells
        if i > 0 and i % 24 == 0:
            c.showPage()

        x, y = coordinates[i % 24]
        create_cell(label, code, x, y, c)

    # save the pdf
    c.save()


if __name__ == "__main__":
    CREATE_CODE = create_datamatrix_code
    
    # stretch barcode,datamatrix not stretched!
    ASPECT_RATIO = True
    

    existing_label_path = os.path.join(__path__, "existing_labels.txt")
    # read existing labels from "existing_labels.txt" file
    with open(existing_label_path, "r") as f:
        existentLabels = [line.strip() for line in f.readlines()]


    thick = [4,8,12]
    sig = {4: [x for x in range(70,150,10)], 8: [x for x in range(70,150,10)], 12: [x for x in range(40,120,10)]}
    bound = ["Z", "A", "B"]
    lfnr = list(range(1,11))
    
    labels = []
    
    for t in thick:
        for s in sig[t]:
            for b in bound:
                for i in lfnr:
                    label = f'{t}.{s}.{b}.{i:02d}'
                    labels.append((label,label))
             
    manual_labels = [
        "SCHOTT1",
        "SCHOTT2",
        "SCHOTT3",
        "SCHOTT4",
        "SCHOTT5",
        "8.140.Z.11",
        "8.140.Z.12",
        "8.140.A.11",
        "8.140.A.12",
    ]
    
    labels = [label for label in labels if label not in existentLabels]

    used_labels = []
    for label in manual_labels:
        used_labels.append((label, label))
    
    for n in range(len(used_labels),24):
        used_labels.append(labels.pop())

    generate_pdf(used_labels, "2023_09_07.pdf")
    
    if os.path.exists(".out"):
        shutil.rmtree(".out", ignore_errors=True)
        
    if os.path.exists("output.pdf"):
        os.system("start output.pdf")

    # append used labels to existing_label_path file
    with open(existing_label_path, "a") as f:
        for label in used_labels:
            f.write(label[0] + "\n")



import os
from pyzbar.pyzbar import decode
from pylibdmtx.pylibdmtx import decode as decode_datamatrix

from matplotlib import pyplot as plt
import cv2
import numpy as np

from fracsuite.splinters.analyzer import isgray

def fourCornersSort(pts):
    """ Sort corners: top-left, bot-left, bot-right, top-right """
    # Difference and sum of x and y value
    # Inspired by http://www.pyimagesearch.com
    diff = np.diff(pts, axis=1)
    summ = pts.sum(axis=1)
    
    # Top-left point has smallest sum...
    # np.argmin() returns INDEX of min
    return np.array([
            pts[np.argmin(summ)],
            pts[np.argmin(diff)],
            pts[np.argmax(summ)],
            pts[np.argmax(diff)],
        ])
                    
def perspective_transform(points,  image_size=(200,200)):
    points = fourCornersSort(points)  # Sort by y-coordinate and then by x-coordinate
    # Define the source points (coordinates of the corners of the original box)
    src_pts = np.array(points, dtype=np.float32)
    
    image_width, image_height = image_size
    
    # Define the destination points (coordinates of the corners of the desired output box)
    # You can specify the dimensions of the output box as per your requirement
    # Here, we'll assume a square box of size 200x200 pixels
    dst_pts = np.array([[0, 0], [image_width, 0], [image_width, image_height], [0, image_height]], dtype=np.float32)
    
    # Calculate the perspective transform matrix
    perspective_matrix = cv2.getPerspectiveTransform(src_pts, dst_pts)
    return perspective_matrix


def display_image(image_array):
    plt.ion()
    
    plt.imshow(image_array)
    plt.show()
    plt.pause(0.1)
    
    text = input("Enter ID: ")
    plt.ioff()
    
    return text

def get_label(image, debug = None):
    orig = image.copy()
    image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    image = cv2.GaussianBlur(image, (15, 15), 20)
    dispImage(image, "LABEL: gaussian blur", "label",debug)
    
    image = cv2.adaptiveThreshold(image, 255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, 21, 1.5)
    dispImage(image, "LABEL: adaptive threshold", "label",debug)
    image = cv2.erode(image, np.ones((3,3), np.uint8), iterations=3)
    dispImage(image, "LABEL: eroded", "label",debug)
    # image = cv2.threshold(image, 127, 255, cv2.THRESH_BINARY)[1]
    # dispImage(image, "LABEL: thresholded", "label",debug)    
    image = cv2.dilate(image, np.ones((3,3), np.uint8), iterations=3)
    dispImage(image, "LABEL: dilated", "label",debug)
    contours, hierarchy = cv2.findContours(255-image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    
    
    
    areas = [cv2.contourArea(x) for x in contours]
    max_i = areas.index(max(areas))
    
    
    x,y,w,h = cv2.boundingRect(contours[max_i])    
    x += 15
    y += 15
    w -= 15
    h -= 15
    
    # create box from xywh
    box = np.array([[x,y],[x+w,y],[x+w,y+h],[x,y+h]])
    
    out = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)    
    dispImage(cv2.drawContours(out, [box], -1, (255,0,0), 10), "LABEL: Label-Contour", "label", debug)
    
    M = perspective_transform(box, (700,370))
    roi = cv2.warpPerspective(orig, M, (700,370))
    
    return roi
    
def find_code(label_original, debug = None):
    well = cv2.cvtColor(label_original, cv2.COLOR_BGRA2GRAY)
    
    # use only the top half of the image
    h,w = well.shape[:2]
    well = well[0:int(h/2), :]
    
    well = cv2.GaussianBlur(well, (5, 5), 1.3)
    dispImage(well, "CODE: Gaussian Blur", "code", debug)
    
    well = cv2.erode(well, np.ones((3,3), np.uint8), iterations=2)
    thr = cv2.threshold(well, 100, 255, cv2.THRESH_BINARY)[1]

    dispImage(thr, "CODE: Thresholded","code", debug)

    contours, hierarchy = cv2.findContours(255-thr, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)    
    areas = [cv2.contourArea(cv2.convexHull(x)) for x in contours]
    max_i = areas.index(max(areas))        

    rect = cv2.minAreaRect(contours[max_i])
    box = cv2.boxPoints(rect)
    box = np.int0(box)      
  
    # extract box from image
    x,y,w,h = cv2.boundingRect(contours[max_i])    
    x -= 15
    y -= 15
    w += 30
    h += 30
    
    dispImage(cv2.drawContours(label_original.copy(), [box], -1, (255,0,0), 10), "CODE: Code-Contour", "code", debug)
    
    # create box from xywh
    box = np.array([[x,y],[x+w,y],[x+w,y+h],[x,y+h]])
    
    roi_image = label_original.copy()

    M = perspective_transform(box, (w,h))
    roi = cv2.warpPerspective(roi_image, M, (w,h))    
    roi = cv2.resize(roi, (int(w*0.5),int(h*0.5)))
    roi = cv2.GaussianBlur(roi, (3, 3), 1.8)
   
    return roi

def read_barcode(image, debug = None):
    """Reads a barcode from an image.

    Args:
        image (numpy.ndarray): Image to read barcode from.

    Returns:
        str: Barcode value.
    """
    
    h,w = image.shape[:2]
    top_right = image[0:int(h/3), int(w/2):]
    dispImage(top_right, "BC: top-right", "barcode", debug)
    label = get_label(top_right, debug)
    dispImage(label, "BC: Label", "barcode", debug)
    code = find_code(label, debug)    
    dispImage(code, "BC: Code", "barcode", debug)
    
    # code = improve_datamatrix(code)
    # dispImage(code)
    
    decoded_objects = decode_datamatrix(code)
    if debug:
        print(decoded_objects)
        
    if decoded_objects:
        for obj in decoded_objects:
            data = obj.data.decode('utf-8')
            return data 
    elif debug:
        dispImage(code, "NOTFOUND: Code", "error", debug)
        
    return display_image(label)



def dispImage(roi, title = "", section = "", debug = None):
    if debug and section not in debug:
        return
    elif not debug:
        return
    
    plt.imshow(roi)
    plt.title(title)
    plt.show()


import os
import argparse
from fracsuite.splinters.analyzer import preprocess_image
from fracsuite.splinters.analyzerConfig import AnalyzerConfig
import numpy as np

from rich import print
import cv2

parser = AnalyzerConfig.get_parser(__doc__)

hsp_group = parser.add_argument_group('High-Speed-Images Arguments')
# hsp_group.add_argument("--dummy", action='store_true')

args = parser.parse_args()

config = AnalyzerConfig.from_args(args)

# get folder
folder = config.path

# get all tiff files from folder and save them
img_files = [os.path.join(folder,x) for x in os.listdir(folder) if x.lower().endswith('tiff')]
print(img_files)

img0 = cv2.imread(img_files[0], cv2.IMREAD_GRAYSCALE)
result_images = []
for i, img in enumerate(img_files[2:]):
    # get two images up to i
    img1 = cv2.imread(img_files[i-1], cv2.IMREAD_GRAYSCALE)
    img2 = cv2.imread(img_files[i], cv2.IMREAD_GRAYSCALE)
    
    img2_o = cv2.cvtColor(img2.copy(), cv2.COLOR_GRAY2BGR)
    
    # transform them first
    # img1 = cv2.subtract(img1, img0)    
    # img2 = cv2.subtract(img2, img0)
    
    # get difference from images
    img_result = cv2.subtract(img1, img2)
    img_result = 255-preprocess_image(img_result, config)
    
    zero = np.zeros_like(img_result)
    img_result = cv2.merge([zero, zero, img_result])
    img_r = cv2.addWeighted(img2_o, 1.0, img_result, 1, 0)
    result_images.append(img_r)
    
    
output_folder = os.path.join(folder, 'out')
os.makedirs(output_folder, exist_ok=True)

for i, im in enumerate(result_images):
    cv2.imwrite(os.path.join(output_folder, f'out_{i:0000}.png'), im)


import argparse
from argparse import RawDescriptionHelpFormatter
import os
from matplotlib import pyplot as plt
from apread import APReader, plot_multiple_datasets, Channel
import numpy as np
from scipy import integrate, signal

descr=\
"""
 █████╗  ██████╗ ██████╗███████╗██╗     ███████╗██████╗  █████╗ ████████╗██╗ ██████╗ ███╗   ██╗
██╔══██╗██╔════╝██╔════╝██╔════╝██║     ██╔════╝██╔══██╗██╔══██╗╚══██╔══╝██║██╔═══██╗████╗  ██║
███████║██║     ██║     █████╗  ██║     █████╗  ██████╔╝███████║   ██║   ██║██║   ██║██╔██╗ ██║
██╔══██║██║     ██║     ██╔══╝  ██║     ██╔══╝  ██╔══██╗██╔══██║   ██║   ██║██║   ██║██║╚██╗██║
██║  ██║╚██████╗╚██████╗███████╗███████╗███████╗██║  ██║██║  ██║   ██║   ██║╚██████╔╝██║ ╚████║
╚═╝  ╚═╝ ╚═════╝ ╚═════╝╚══════╝╚══════╝╚══════╝╚═╝  ╚═╝╚═╝  ╚═╝   ╚═╝   ╚═╝ ╚═════╝ ╚═╝  ╚═══╝
                                                                                               
Leon Bohmann            TUD - ISMD - GCC              www.tu-darmstadt.de/glass-cc

Description:
-------------------------
This module is used to analyze acceleration data, measured during the fracturing of the
glass plys. The sensor data is read using the apreader from https://github.com/leonbohmann/apreader.
  

Usage:
-------------------------
Command line usage is shown below. For further information visit:
https://github.com/leonbohmann/fracture-suite
"""  

def time_to_index(time_array, t):
    """Returns the index of the time array that is closest to the given time t."""
    return np.argmin(np.abs(time_array - t))

def plotChannels(chans: list[Channel], lbs: list[str], title: str):
    styles = [None] * len(chans)
    for x in range(0,len(chans)):
        if len(lbs) > x:
            styles[x] = lbs[x]
            
    return plot_multiple_datasets([(x.Time.data, x.data, l, f'{x.Name} [{x.unit}]', x.Name) for x,l in zip(chans, styles)], title)


def test_1(reader: APReader):
    time = reader.Groups[0].ChannelX.data
    # impact_time_id = np.argmax([])    
    # time = time - time[impact_time_id-5]
    print([x.Name for x in reader.Channels])

    ds = reader.collectChannels(['Force', 'Acc1', 'Fall_g2'])

    fig,axs = plot_multiple_datasets([ \
        (time, ds[0].data, "g-", "Force [N]", "Fall 2"),
        (time, ds[1].data, "r-", "PCB Acc [g]", "PCB Acc"),
        (time, ds[2].data, "b-", "Endevco Acc[g]", "Endevco Acc")], 
        'Acceleration delay in fall weight')
    
    fig.savefig(os.path.join(out_dir, f"{reader.fileName}_detail.png"))
    
def test_4(reader: APReader):
    time = reader.Groups[0].ChannelX.data
    # impact_time_id = np.argmax([])    
    # time = time - time[impact_time_id-5]
    print([x.Name for x in reader.Channels])

    ds = reader.collectChannels(['Force', 'Fall_g2'])

    fig,axs = plot_multiple_datasets([ \
        (time, ds[0].data, "g-", "Force [N]", "Force"),
        (time, ds[1].data, "r-", "Fall_g2 [g]", "Fall_g2")], 
        'Acceleration delay in fall weight')
    
    fig.savefig(os.path.join(out_dir, f"{reader.fileName}_detail.png"))

def test_hammer_sensor_fallgewicht(reader: APReader):
    time = reader.Groups[0].ChannelX.data
    # impact_time_id = np.argmax([])    
    # time = time - time[impact_time_id-5]

    ds = reader.collectChannels(['Force', 'Acc1', 'Fall_g2'])

    fig,axs = plotChannels(ds, ['r-', 'g-', 'b-'], 
        'Acceleration delay in fall weight')
    
    fig.savefig(os.path.join(out_dir, f"{reader.fileName}_detail.png"))


def test_impact_delay(reader:APReader):
    ds = reader.collectChannels(['Fall_g1', 'Fall_g2', 'Acc2', 'Acc6', 'Acc3'])

    fig,axs = plotChannels(ds, ['b--', 'g--'], 
        'Acceleration delay in fall weight')
    
    fig.savefig(os.path.join(out_dir, f"{reader.fileName}_impact_delay.png"))

def test_3(reader: APReader):
    time = reader.Groups[0].ChannelX.data
    # impact_time_id = np.argmax([])    
    # time = time - time[impact_time_id-5]
    print([x.Name for x in reader.Channels])

    ds = reader.collectChannels(['Force', 'Acc1', 'Acc2', 'Acc3', 'Acc4', 'Acc7'])

    fig,axs = plot_multiple_datasets([ \
        (time, ds[0].data, "g-", "Force [N]", "Fall 2"),
        (time, ds[1].data, "r-", "Acc 1 [g]", "PCB Acc"),
        (time, ds[2].data, "k-", "Acc 2 (S) [g]", "PCB Acc"),
        (time, ds[3].data, "y-", "Acc 3 [g]", "PCB Acc"),
        (time, ds[4].data, "m-", "Acc 4 (S) [g]", "PCB Acc"),
        (time, ds[5].data, "b-", "Acc 7 [g]", "Acc 7")], 
        'Acceleration delay in fall weight')
    
    fig.savefig(os.path.join(out_dir, f"{reader.fileName}_detail.png"))


def test_2(reader: APReader):
    channel_max = [(x.Name, np.max(x.data), np.min(x.data)) for x in reader.Channels]

    for max in channel_max:
        print(f'{max[0]}: {max[1]} / {max[2]}')

    # channels: 0: Piezoelectric, 1: Piezoresistive
    channels = [x for x in reader.Channels if x.Name.startswith("Fall")]

    g = 9.81 # m/s²
    drop_acc = channels[1] 
    drop_acc1 = channels[0].data
    time = drop_acc.Time.data

    # this shifts the acc signal to zero before the impact
    drop_avg = np.average(drop_acc.data[-120000:])
    drop_data = drop_acc.data - drop_avg
    drop_data = drop_data * g

    # estimate savgol filter parameters based on the data time interval
    time_interval = time[1] - time[0]
    window_length = int(0.1 / time_interval)
    if window_length % 2 == 0:
        window_length += 1
    polyorder = 3


    # savgol
    drop_data_smooth = signal.savgol_filter(drop_data, 140, 5)

    # firwin
    cutoff_frequency = 0.1  # Adjust as needed
    num_taps = 101  # Filter length, adjust as needed
    fir_filter = signal.firwin(num_taps, cutoff_frequency)
    # drop_data_smooth = signal.lfilter(fir_filter, 1.0, drop_data)
    # np.savetxt("output.txt", np.column_stack([time, drop_data]), fmt = ['%.8f', '%.8f'])
    # drop_data = drop_data_smooth

    a = drop_data.copy()

    impact_w = 10
    impact_time_id = np.argmax(np.abs(a))
    impact_end_index = impact_time_id + impact_w//2
    impact_start_index = impact_time_id - impact_w//2

    # smooth g around the impact
    a[:impact_start_index] = drop_data_smooth[:impact_start_index]
    a[impact_end_index:] = drop_data_smooth[impact_end_index:]


    # shift a to zero around the impact
    impact_time = time[impact_time_id]
    impact_start_index = time_to_index(time, impact_time - 0.5)
    impact_end_index = time_to_index(time, impact_time + 1.5)


    # take aaverage around impact_start and end index

    # shift g into zero
    g_shift = np.average(a[:impact_start_index])
    g_shift2 = np.average(a[impact_end_index:])
    g_shift =( g_shift + g_shift2) / 2
    a = a - g_shift


    # initial is needed, so that the length of the result is the same as the input
    v = integrate.cumulative_trapezoid(a, time, initial = 0)
    s = integrate.cumulative_trapezoid(v, time, initial = 0)

    # shift the drop_data to zero after the impact
    # find the impact time index
    # impact_time_id = np.argmax(np.abs(v))
    # impact_time = time[impact_time_id]
    # impact_end_index = time_to_index(time, impact_time + 1.5)

    # g_shift = np.average(drop_data_smooth[impact_end_index:])
    # drop_data_s = drop_data_smooth.copy()
    # # shift data from impact to end
    # drop_data_s[impact_time_id+1:] = drop_data_s[impact_time_id+1:] - g_shift

    # # calculate g shift after impact
    # v_shift = np.average(v[impact_end_index:])
    # v_shifted = v.copy()
    # # shift data from impact to end
    # v_shifted[impact_time_id+1:] = v_shifted[impact_time_id+1:] - v_shift

    # use the corrected g now
    # v = integrate.cumulative_trapezoid(drop_data_s, time, initial = 0)

    # s_shifted = integrate.cumulative_trapezoid(v_shifted, time, initial = 0)
    # s = integrate.cumulative_trapezoid(v, time, initial = 0)

    # fig0 = plt.plot()

    # plt.plot(time, drop_data, "gray", label="Acceleration")
    # plt.plot(time, drop_data_smooth, "b", label="Smoothed")
    # plt.plot(time, drop_data_s, "g", label="Adjusted")

    # plt.legend()
    # plt.show()

    fig,axs = plot_multiple_datasets([ \
        (time, drop_acc1, "k--", "Acc 1 [g]", "Ac1"), \
        (time, a, "m-", "Acc [m/s²]", "Acceleration"),
        (time, drop_data, "r--", "Acc [m/s²]", "Acceleration"),
        # (time, v_shifted, "b--", "Speed [m/s]", "Acceleration"), 
        (time, v, "b", "Speed [m/s]", "Speed"),
        # (time, s_shifted, "g--", "Distance [m]", "Distance"),
        (time, s, "g", "Distance [m]", "Distance")], 
        'Time integrals of Acceleration')
    fig.savefig(os.path.join(out_dir, f"{reader.fileName}_fall.png"))
    
    time = time - time[impact_time_id-5]
    print([x.Name for x in reader.Channels])
    # Auswertung auf Impact beziehen und Zeit ab dort messen
    acc1 = reader.Channels[1]
    acc2 = reader.Channels[2]
    acc3 = reader.Channels[3]
    acc4 = reader.Channels[4]
    acc5 = reader.Channels[5]
    acc6 = reader.Channels[6]
    acc_g1 = reader.Channels[7]
    acc_g2 = reader.Channels[8]


    fig,axs = plot_multiple_datasets([ \
        # (time, drop_acc1, "k-", "Fall 1 [g]", "Fall 1"), \
        (time, acc_g2.data, "g-", "Fall 2 [g]", "Fall 2"), \
        (time, acc3.data, "r-", "Acc 3 [g]", "Acc 2"), \
        (time, acc2.data, "b-", "Acc 2 [g]", "Acc 2"), \
        (time, acc6.data, "y-", "Acc 6 [g]", "Acc 6")], 
        'Comparison of different impact times')
    fig.savefig(os.path.join(out_dir, f"{reader.fileName}_fall2.png"))

# get all test-functions
test_funcs = [x for x in globals().values() if callable(x) and x.__name__.startswith('test')]
# strip leading underscore and test from names
test_names = sorted([x.__name__[5:] for x in test_funcs])

parser = argparse.ArgumentParser(description=descr, formatter_class=RawDescriptionHelpFormatter)    

parser.add_argument('measurement', nargs="?", \
    help="""The measurement to be processed. This can either be a .bin file or a project folder that
    has subfolders 'fracture/acc', where the bin file is located.""")
parser.add_argument('-test', nargs=1, default=None, choices=test_names,
    help="""Run the test environment""")
parser.add_argument('--sameaxis', action="store_true", help='Plot all datasets in the same axis.')

args = parser.parse_args()

if args.measurement.lower().endswith('.bin'):
    file = args.measurement
else:
    search_path = os.path.join(args.measurement, 'fracture', 'acceleration')
    for pfile in os.listdir(search_path):
        if pfile.lower().endswith('bin'):
            file = os.path.join(search_path, pfile)
            break

out_dir = os.path.dirname(file)


# file = r"d:\Forschung\Glasbruch\Versuche.Reihe\Proben\4.70.30.B\fracture\acc\4.70.30.A_fracture.bin"
# pfile = r"d:\Forschung\Glasbruch\Versuche.Reihe\Proben\8.140.Z.2\frac\8_140_Z_2.bin"
reader = APReader(file, verbose=True)

reader.printSummary()
reader.plot(sameAxis=args.sameaxis)

if args.test is not None:
    fname = f'test_{args.test[0]}'
    print(f'Calling {fname}')
    func = globals()[fname]
    func(reader)

# slow_group = reader.Groups[0]
# fall_group = reader.collectDatasets(["Fall_g1", "Fall_g2"])

# plot_multiple_datasets(fall_group, 'Fallgewicht')

# slow_group.plot()






from __future__ import annotations
import csv

import os
import random

import cv2
import numpy as np
import numpy.typing as nptyp
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from matplotlib.widgets import RectangleSelector
import matplotlib.colors as colors
import matplotlib.cm as cmx
from scipy.spatial import Voronoi, voronoi_plot_2d
from scipy.spatial.distance import pdist
from skimage.morphology import skeletonize
from tqdm import tqdm
from rich import inspect, print
from fracsuite.splinters.analyzerConfig import AnalyzerConfig
from fracsuite.splinters.splinter import Splinter

from scipy.spatial import Delaunay

def isgray(img):
        if len(img.shape) < 3:
            return True
        if img.shape[2] == 1:
            return True
        # b,g,r = img[:,:,0], img[:,:,1], img[:,:,2]
        return False

def plotImage(img,title:str, color: bool = True, region: tuple[int,int,int,int] = None):
    if isgray(img) and color:
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)

    fig, axs = plt.subplots()
    axs.imshow(img)
    axs.set_title(title)

    if region is not None:
        (x1, y1, x2, y2) = region
        axs.set_xlim((x1,x2))
        axs.set_ylim((y1,y2))

    plt.show()


def crop_perspective(img, cropped_image_size: tuple[int,int], debug: bool):
    """
    Crops a given image to its containing pane bounds. Finds smallest pane countour with
    4 corner points and aligns, rotates and scales the pane to fit a resulting image.

    Args:
        img (Image): Input image with a clearly visible glass pane.
        size (int): Size of the resulting image after perspective transformation.
        dbg (bool): Displays debug image outputs during processing.

    Returns:
        img: A cropped image which only contains the glass pane. Size: 1000x1000.
            If no 4 point contour is found, the whole image is returned.
    """

    def fourCornersSort(pts):
        """ Sort corners: top-left, bot-left, bot-right, top-right """
        # Difference and sum of x and y value
        # Inspired by http://www.pyimagesearch.com
        diff = np.diff(pts, axis=1)
        summ = pts.sum(axis=1)

        # Top-left point has smallest sum...
        # np.argmin() returns INDEX of min
        return np.array([pts[np.argmin(summ)],
                        pts[np.argmax(diff)],
                        pts[np.argmax(summ)],
                        pts[np.argmin(diff)]])



    img_original = img.copy()

    im = img.copy()
    im0 = img.copy()

    if not isgray(im):
        im = cv2.cvtColor(img,cv2.COLOR_BGR2GRAY)

    if isgray(im0):
        im0 = cv2.cvtColor(im0,cv2.COLOR_GRAY2BGR)


    # # apply gaussian blur to image to we can get rid of some noise
    im = cv2.GaussianBlur(im, (5,5), 5)
    # # restore original image by thresholding
    _,im = cv2.threshold(im,127,255,0)

    if debug:
        plotImage(im, 'CROP: Image for rectangle detection')

    # fetch contour information
    contour_info = []
    contours, _ = cv2.findContours(255-im, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    # loop through contours and find their properties
    for cnt in contours:
        contour_info.append((
            cnt,
            cv2.isContourConvex(cnt),
            cv2.contourArea(cnt)
        ))

    # os.makedirs('.debug', exist_ok=True)
    # # count files in debug folder
    # file_count = len(os.listdir('.debug'))

    # cv2.imwrite(os.path.join('.debug', f'contours_{file_count}.png'),
    #             cv2.drawContours(im0, contours, -1, (0,0,255), 10))

    # sort contours after their area
    contour_info = sorted(contour_info, key=lambda c: c[2], reverse=True)
    # take the second largest contour (this has to be the outer bounds of pane)
    if len(contour_info) > 0:
        max_contour = contour_info[0][0]
    else:
        return img_original

    cv2.drawContours(im0, contours, -1, (0,0,255), 10)
    if debug:
        plotImage(im0, 'CROP: Detected contours')

    # Simplify contour
    perimeter = cv2.arcLength(max_contour, True)
    approx = cv2.approxPolyDP(max_contour, 0.03 * perimeter, True)

    # Page has 4 corners and it is convex
    # Page area must be bigger than maxAreaFound
    if (len(approx) == 4 and
            cv2.isContourConvex(approx)):
        pageContour = fourCornersSort(approx[:, 0])

    else:
        rect = cv2.boundingRect(approx)
        x, y, w, h = rect

        # Compute the four corners of the rectangle using cv2.boxPoints
        corners = cv2.boxPoints(((x, y), (w, h), 0))

        # Convert the corners to integer values and print the result
        corners = corners.astype(int)
        #raise CropException("Pane boundary could not be found.")
        pageContour = corners

    if debug:
        cv2.drawContours(im0, [pageContour],-1, (0,0,255), thickness=im.shape[0]//50)
        plotImage(im0, 'CROP: Found rectangle contour')



    im_h, im_w =im.shape[0],im.shape[1]

    if im_w > im_h:
        f = im_w / 4000
        im_w = 4000
        im_h = int(im_w / f)
    else:
        f = im_h / 4000
        im_h = 4000
        im_w = int(im_h / f)

    # Create target points
    if cropped_image_size is not None:
        width, height=cropped_image_size
    else:
        height,width=im_w, im_h

    tPoints = np.array([[0, 0],
                    [0, height],
                    [width, height],
                    [width, 0]], np.float32)
    # source points are contour corners
    sPoints = pageContour

    # getPerspectiveTransform() needs float32
    if sPoints.dtype != np.float32:
        sPoints = sPoints.astype(np.float32)
    # Warping perspective
    M = cv2.getPerspectiveTransform(sPoints, tPoints)

    img_original = cv2.warpPerspective(img_original, M, (int(width), int(height)))
    # print(img_original.shape)
    # plotImage(img_original, 'CROP: Cropped image')
    # and return the transformed image
    return img_original

def to_rgb(img):
    if isgray(img):
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    return img


def filter_contours(contours, hierarchy, config: AnalyzerConfig) \
    -> list[nptyp.ArrayLike]:
    """
    This function filters a list of contours.
    """
    if config.debug:
        len_0 = len(contours)
    # Sort the contours by area (desc)
    contours, hierarchy = zip(*sorted(zip(contours, hierarchy[0]), \
        key=lambda x: cv2.contourArea(x[0]), reverse=True))

    contours = list(contours)
    contours_areas = [cv2.contourArea(x) for x in contours]
    # contour_area_avg = np.average(contours_areas)
    # contour_area_med = np.average(contours_areas)
    # contour_area_std = np.std(contours_areas)

    # Create a list to store the indices of the contours to be deleted
    to_delete: list[int] = []
    As = []
    ks = []
    # Iterate over all contours
    for i in tqdm(range(len(contours)), desc = 'Eliminate small contours', leave=False):
        contour_area = contours_areas[i]
        contour_perim = cv2.arcLength(contours[i], False)


        if contour_area > 0:
            As.append(contour_area)
            ks.append(contour_perim / contour_area)

        if contour_area == 0:
            contour_area = 0.00001

        # small size
        if contour_perim < config.fragment_min_area_px:
            to_delete.append(i)

    # def reject_outliers(data, m = 2.):
    #     return data[abs(data - np.mean(data)) < m * np.std(data)]

        # filter outliers
        elif contour_area > config.fragment_max_area_px:
            to_delete.append(i)

        # # more line shaped contour
        # elif contour_perim / contour_area > 1:
        #     contours[i] = cv2.convexHull(contours[i])

        # # Check if the contour has a parent
        # elif hierarchy[i][1] != -1:
        #     # If so, mark the contour for deletion
        #     to_delete.append(i)

        # # all other cases
        # else:
        #     contours[i] = connect_closest_hard_angles(contours[i], 140)
        #     contours[i] = cv2.approxPolyDP(contours[i], 0.1, True)

    # Delete the marked contours
    for index in sorted(to_delete, reverse=True):
        del contours[index]

    if config.debug:
        len_1 = len(contours)
        print(f'FILT: Contours before: {len_0}, vs after: {len_1}')

    return contours

def csintkern(events, region, h):
    n, d = events.shape

    # Get the ranges for x and y
    minx = np.min(region[:][0])
    maxx = np.max(region[:][0])
    miny = np.min(region[:][1])
    maxy = np.max(region[:][1])

    # Get 50 linearly spaced points
    xd = np.linspace(minx, maxx, 50)
    yd = np.linspace(miny, maxy, 50)
    X, Y = np.meshgrid(xd, yd)
    st = np.column_stack((X.ravel(), Y.ravel()))
    ns = len(st)
    xt = np.vstack(([0, 0], events))
    z = np.zeros(X.ravel().shape)

    for i in tqdm(range(ns), leave=False):
        # for each point location, s, find the distances
        # that are less than h.
        xt[0] = st[i]
        # find the distances. First n points in dist
        # are the distances between the point s and the
        # n event locations.
        dist = pdist(xt)
        ind = np.where(dist[:n] <= h)[0]
        t = (1 - dist[ind]**2 / h**2)**2
        z[i] = np.sum(t)

    z = z * 3 / (np.pi * h)
    Z = z.reshape(X.shape)

    return X, Y, Z

def csintkern_optimized(events, region, h=500):
    n, d = events.shape

    # Get the ranges for x and y
    minx = np.min(region[:][0])
    maxx = np.max(region[:][0])
    miny = np.min(region[:][1])
    maxy = np.max(region[:][1])

    # Get 50 linearly spaced points
    xd = np.linspace(minx, maxx, 50)
    yd = np.linspace(miny, maxy, 50)
    X, Y = np.meshgrid(xd, yd)
    st = np.column_stack((X.ravel(), Y.ravel()))
    ns = len(st)

    # Precompute distances
    dist_matrix = np.linalg.norm(events[:, np.newaxis, :] - st[np.newaxis, :, :], axis=-1)
    t = np.maximum(0, 1 - (dist_matrix / h) ** 2) ** 2

    # Calculate intensities
    z = np.sum(t, axis=0)
    z *= 3 / (np.pi * h)

    Z = z.reshape(X.shape)

    return X, Y, Z

def detect_fragments(binary_image, config: AnalyzerConfig) -> list[nptyp.ArrayLike]:
    try:
        # Find contours of objects in the binary image
        contours, hierar = \
            cv2.findContours(binary_image, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        contours = list(contours)
        contours = filter_contours(contours, hierar, config)
        return contours
    except Exception as e:
        raise ValueError(f"Error in fragment detection: {e}")

def closeImg(image, sz = 3, it=1):
    kernel_closing = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (sz, sz))
    return cv2.morphologyEx(image, cv2.MORPH_CLOSE, kernel_closing, iterations=it)

def rand_col():
    """Generate a random color.

    Returns:
        (r,g,b): A random color.
    """
    return (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))

def get_color(value, min_value = 0, max_value = 1, colormap_name='viridis'):
    # Normalize the value to be in the range [0, 1]
    normalized_value = (value - min_value) / (max_value - min_value)

    # Choose the colormap
    colormap = plt.get_cmap(colormap_name)

    # Map the normalized value to a color
    color = colormap(normalized_value)

    # Convert the RGBA color to RGB
    rgb_color = colors.to_rgba(color)[:3]

    return tuple(int(255 * channel) for channel in rgb_color)

def preprocess_spot_detect(img):
    if not isgray(img):
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    img = cv2.GaussianBlur(img, (5,5), 3)
    img = cv2.threshold(img, 100, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)[1]
    return img


class Analyzer(object):
    """
    Analyzer class that can handle an input image.
    """

    file_path: str
    "Path of the input image."
    file_dir: str = None
    "Directory of the input image."
    out_dir: str
    "Directory of the output images."

    original_image: nptyp.ArrayLike
    "Original image, after cropping (if applied)."
    preprocessed_image: nptyp.ArrayLike
    "Preprocessed image, after cropping (if applied) and pre-processing."

    image_contours: nptyp.ArrayLike
    "Image with all contours drawn."
    image_filled: nptyp.ArrayLike
    "Image with all contours filled."

    contours: list[nptyp.ArrayLike]
    "List of all contours."
    splinters: list[Splinter]
    "List of all splinters."

    axs: list[plt.Axes]
    "List of all axes."
    fig_comparison: Figure
    fig_area_distr: Figure
    fig_area_sum: Figure

    config: AnalyzerConfig
    "Configuration of the analyzer."

    def __init__(self, config: AnalyzerConfig = None):
        """Create a new analyzer object.

        Args:
            file_path (str): Path to image.
            crop (bool, optional): Specify, if the input has to be cropped first.
                Defaults to False.
            img_size (int, optional): Size of the cropped image.
                Defaults to 4000, only needed if crop=True.
            img_real_size (tuple[int,int], optional):
                Actual size in mm of the input rectangular ply.
        """
        if config is None:
            config = AnalyzerConfig()

        self.config = config

        #############
        # folder operations
        if config.path.endswith('\\'):
            search_path = os.path.join(config.path, 'fracture', 'morphology')
            for file in os.listdir(search_path):
                if 'Transmission' in file and file.endswith('.bmp'):
                    print(f"[green]Found image in specimen folder.[/green]")
                    self.file_path = os.path.join(search_path, file)
                    self.file_dir = os.path.dirname(self.file_path)
                    break
            self.out_dir = os.path.join(config.path, 'fracture', 'splinter')

            if self.file_dir is None:
                raise Exception("Could not find a morphology file in the specified folder.")
        else:
            self.file_path = config.path
            self.file_dir = os.path.dirname(config.path)
            self.out_dir = os.path.join(self.file_dir, config.out_name)

            self.out_dir = os.path.join(self.out_dir, os.path.splitext(os.path.basename(config.path))[0])


        print(f"Input file: '[bold]{self.file_path}[/bold]'")
        print(f"Output directory: '[bold]{self.out_dir}[/bold]'")
        # create output directory if not exists
        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)

        #############
        # image operations
        print('> Step 1: Preprocessing image...')
        self.original_image = cv2.imread(self.file_path, cv2.IMREAD_GRAYSCALE)
        if config.crop:
            self.original_image = crop_perspective(self.original_image, config.cropped_image_size, config.debug)
            # this is the default for the input images from cullet scanner
            self.original_image = cv2.rotate(self.original_image, cv2.ROTATE_90_COUNTERCLOCKWISE)
            self.original_image = cv2.cvtColor(self.original_image, cv2.COLOR_GRAY2BGR)

        self.preprocessed_image = self.preprocess_image(self.original_image, config)



        #############
        # calculate scale factors to make measurements in real units
        size_f = 1
        "size factor for mm/px"
        if config.real_image_size is not None and config.cropped_image_size is not None:
            # fx: mm/px
            fx = config.real_image_size[0] / config.cropped_image_size[0]
            fy = config.real_image_size[1] / config.cropped_image_size[1]

            # the factors must match because pixels are always squared
            # for landscape image, the img_real_size's aspect ratio must match
            if fx != fy:
                raise Exception("The scale factors for x and y must match!" + \
                    "Check the input image and also the img_real_size parameter.")

            # f: mm/px
            size_f = fx
            config.size_factor = size_f

        #############
        # initial contour operations
        print('> Step 2: Analyzing contours...')
        all_contours = detect_fragments(self.preprocessed_image, config)
        stencil = np.zeros((self.preprocessed_image.shape[0], \
            self.preprocessed_image.shape[1]), dtype=np.uint8)

        for c in all_contours:
            cv2.drawContours(stencil, [c], -1, 255, thickness = -1)

        #############
        # advanced image operations
        # first step is to skeletonize the stencil
        print('> Step 2.1: Skeletonize 1|2')
        skeleton = skeletonize(255-stencil)
        skeleton = skeleton.astype(np.uint8) * 255
        cv2.imwrite(self.__get_out_file('.debug/skeleton.png'), skeleton)

        if config.debug:
            plotImage(skeleton, 'SKEL1', False, region=config.display_region)
        skeleton = closeImg(skeleton, config.skelclose_size, config.skelclose_amnt)

        cv2.imwrite(self.__get_out_file('.debug/closed.png'), skeleton)

        if config.debug:
            plotImage(skeleton, 'SKEL1 - Closed', False, region=config.display_region)
        # second step is to skeletonize the closed skeleton from #1
        print('> Step 2.2: Skeletonize 2|2')
        skeleton = skeletonize(skeleton)
        skeleton = skeleton.astype(np.uint8) * 255
        cv2.imwrite(self.__get_out_file('.debug/closed_skeleton.png'), skeleton)
        if config.debug:
            plotImage(skeleton, 'SKEL2', False, region=config.display_region)

        self.image_skeleton = skeleton.copy()

        #############
        # detect fragments on the closed skeleton
        print('> Step 3: Preliminary contour analysis...')
        self.contours = detect_fragments(skeleton, config)
        self.splinters = [Splinter(x,i,size_f) for i,x in enumerate(self.contours)]

        #############
        # filter dark spots and delete fragments
        if not config.skip_darkspot_removal:
            print('> Step 4: Filter spots...')
            self.__filter_dark_spots(config)
        else:
            print('> Step 4: Filter spots (SKIPPED)')

        #############
        # detect fragments on the closed skeleton
        print('> Step 5: Final contour analysis...')
        self.contours = detect_fragments(self.image_skeleton, config)
        self.splinters = [Splinter(x,i,size_f, config) for i,x in enumerate(self.contours)]

        self.__save_data(config)

        self.image_skeleton_rgb = cv2.cvtColor(skeleton, cv2.COLOR_GRAY2BGR)


        #############
        # create images
        self.image_contours = self.original_image.copy()
        self.image_filled = self.original_image.copy()
        for c in self.contours:
            cv2.drawContours(self.image_contours, [c], -1, rand_col(), 1)
        for c in self.contours:
            cv2.drawContours(self.image_filled, [c], -1, rand_col(), -1)
        cv2.imwrite(self.__get_out_file(f"img_filled.{self.config.ext_imgs}"), self.image_filled)
        cv2.imwrite(self.__get_out_file(f"img_contours.{self.config.ext_imgs}"), self.image_contours)
        cv2.imwrite(self.__get_out_file(f"debug_preprocessed.{self.config.ext_imgs}"), self.preprocessed_image)
        cv2.imwrite(self.__get_out_file(f"debug_skeleton.{self.config.ext_imgs}"), self.image_skeleton_rgb)

        combined = cv2.addWeighted(self.image_contours, 1.0, self.image_filled, 0.3, 0.0)
        cv2.imwrite(self.__get_out_file(f"img_contours_filled_combined.{self.config.ext_imgs}"), combined)


        #############
        # Stochastic analysis
        print("> Step 6: Stochastic analysis...")
        self.__create_voronoi(config)

        #############
        # Orientational analysis
        if config.impact_position is not None:
            print("> Step 7: Orientation analysis...")
            self.__create_impact_influence(size_f, config)

        #############
        # count splinters in norm region
        if config.norm_region_center is not None:
            print("> Step 8: Count splinters in norm region...")
            self.__count_splinters_in_norm_region(config)
        else:
            print("> Step 8: (SKIPPED)")


        self.__create_splintersize_filled_image(config)

        print("> Create output plots...")
        self.__plot_backend(config.display_region, display=config.displayplots)
        self.__plot_splintersize_accumulation(display=config.displayplots)
        self.__plot_splintersize_distribution(display=config.displayplots)

        ############
        ############
        # Summary
        print('\n\n')
        print(f'Splinter count: {len(self.contours)}')
        self.__check_detection_ratio(config, doprint=True)
        print('\n')

    def preprocess_image(self, image, config: AnalyzerConfig) -> nptyp.ArrayLike:
        """Preprocess a raw image.

        Args:
            image (nd.array): The input image.
            gauss_sz (size-tuple, optional): The size of the gaussian filter.
                                            Defaults to (5,5).
            gauss_sig (int, optional): The sigma for gaussian filter. Defaults to 3.
            block_size (int, optional): Block size of adaptive threshold. Defaults to 11.
            C (int, optional): Sensitivity of adaptive threshold. Defaults to 6.
            rsz (int, optional): Resize factor for the image. Defaults to 1.

        Returns:
            np.array: Preprocessed image.
        """

        rsz_fac = config.resize_factor # x times smaller

        if not isgray(image):
            image = cv2.cvtColor(image.copy(), cv2.COLOR_BGR2GRAY)

        # Apply Gaussian blur to reduce noise and enhance edge detection
        image = cv2.GaussianBlur(image, config.gauss_size, config.gauss_sigma)
        image = cv2.resize(image,(int(image.shape[1]/rsz_fac), int(image.shape[0]/rsz_fac)))

        cv2.imwrite(self.__get_out_file('.debug/step1.png'), image)
        if config.debug:
            plotImage(image, 'PREP: GaussianBlur -> Resize', region=config.display_region)

        # Use adaptive thresholding
        image = cv2.adaptiveThreshold(image, 255, cv2.ADAPTIVE_THRESH_MEAN_C, \
            cv2.THRESH_BINARY, config.thresh_block_size, config.thresh_sensitivity)

        cv2.imwrite(self.__get_out_file('.debug/step2.png'), image)
        if config.debug:
            plotImage(image, 'PREP: ... -> Adaptive Thresh', region=config.display_region)

        return image

    def __save_data(
        self,
        config: AnalyzerConfig,
        splinters: list[Splinter] = None,
        fname: str = 'splinter_data.csv'

    ) -> None:
        """Save splinter data to a csv file.

        Args:
            config (AnalyzerConfig): Configuration.
        """

        if splinters is None:
            splinters = self.splinters

        # save data to csv file
        with open(self.__get_out_file(fname), 'w', newline='') as csvfile:
            writer = csv.writer(csvfile, delimiter=';')
            writer.writerow(['id', 'area', 'circumfence', 'alignment_score', 'angle', 'roughness', 'roundness', 'c_x [px]', 'c_y [px]', 'c_x [mm]', 'c_y [mm]'])
            for i, s in tqdm(enumerate(splinters), desc='Saving data...', leave=False):
                writer.writerow([i, s.area, s.circumfence, s.alignment_score, s.angle, s.roughness, s.roundness, s.centroid_px[0], s.centroid_px[1], s.centroid_mm[0], s.centroid_mm[1]])

    def __count_splinters_in_norm_region(self, config: AnalyzerConfig) -> float:
        # create rectangle around args.normregioncenter with 5x5cm size
        # and count splinters in it
        x,y = config.norm_region_center
        w,h = config.norm_region_size
        x1 = x - w // 2
        x2 = x + w // 2
        y1 = y - h // 2
        y2 = y + h // 2
        # transform to real image size
        x1 = int(x1 // config.size_factor)
        x2 = int(x2 // config.size_factor)
        y1 = int(y1 // config.size_factor)
        y2 = int(y2 // config.size_factor)

        s_count = 0
        splinters_in_region: list[Splinter] = []
        # count splinters in norm region
        for s in self.splinters:
            if s.in_region_exact((x1,y1,x2,y2)):
                s_count += 1
                splinters_in_region.append(s)

        print(f'Splinters in norm region: {s_count}')

        # plot splinter PDF
        fig, axs = plt.subplots()
        axs.hist([x.area for x in splinters_in_region], bins=25)
        axs.set_xlabel('Splinter area [mm²]')
        axs.set_ylabel('PDF P(A)')
        fig.savefig(self.__get_out_file(f"norm_count_pdf.{config.ext_plots}"))
        self.__save_data(config, splinters_in_region, 'norm_count_splinters.csv')

        norm_filled_img = np.zeros((self.original_image.shape[0], self.original_image.shape[1], 3), dtype=np.uint8)
        print(norm_filled_img.shape)
        for s in splinters_in_region:
            clr = rand_col()
            cv2.drawContours(norm_filled_img, [s.contour], -1, clr, -1)


        # # get norm region from original image (has to be grayscale for masking)
        # norm_region_mask = np.zeros(self.original_image.shape[:2], dtype=np.uint8)
        # cv2.rectangle(norm_region_mask, (x1,y1), (x2,y2), 255, -1)
        # # create image parts
        # normed_image = cv2.bitwise_and(norm_filled_img, norm_filled_img, mask=norm_region_mask)
        # normed_image_surr = self.original_image #cv2.bitwise_and(self.original_image, self.original_image, mask=norm_region_inv)
        # # add images together
        normed_image = cv2.addWeighted(to_rgb(self.original_image), 1, norm_filled_img, 0.5, 0)
        cv2.rectangle(normed_image, (x1,y1), (x2,y2), (255,0,0), 5)

        # extract image part
        norm_region_image = normed_image[y1-50:y2+50, x1-50:x2+50]
        cv2.imwrite(self.__get_out_file(f"norm_region_image.{config.ext_imgs}"), norm_region_image)

        # extract overview image
        cv2.putText(normed_image, f'{s_count}', (x1,y1), cv2.FONT_HERSHEY_SIMPLEX, 6, (0,0,255), 20)
        cv2.imwrite(self.__get_out_file(f"norm_count.{config.ext_imgs}"), cv2.resize(normed_image, (0,0), fx=0.5, fy=0.5))




        return s_count

    def __check_detection_ratio(self, config: AnalyzerConfig, doprint = False) -> float:
        #############
        # check percentage of detected splinters
        total_area = np.sum([x.area for x in self.splinters])

        if config.real_image_size is not None:
            total_img_size = config.real_image_size[0] * config.real_image_size[1]
        else:
            total_img_size = self.preprocessed_image.shape[0] * \
                self.preprocessed_image.shape[1]

        p = total_area / total_img_size * 100
        if doprint:
            print(f'Detection Ratio: {p:.2f}%')

        return p

    def __create_impact_influence(self, size_f: float, config: AnalyzerConfig):
        # analyze splinter orientations
        orientation_image = self.original_image.copy()
        orients = []
        for s in tqdm(self.splinters, leave=False, desc='Analyzing orientations...'):
            orientation = s.measure_orientation(config)
            orients.append(orientation)
            color = get_color(orientation, colormap_name='turbo')
            cv2.drawContours(orientation_image, [s.contour], -1, color, -1)
            # p2 = (s.centroid_px + s.angle_vector * 15).astype(np.int32)
            # cv2.line(orientation_image, s.centroid_px, p2, (255,255,255), 3)
        cv2.circle(orientation_image, (np.array(config.impact_position) / size_f).astype(np.uint32),
                   np.min(orientation_image.shape[:2]) // 50, (255,0,0), -1)

        # save plot
        fig, axs = plt.subplots()
        axim = axs.imshow(orientation_image, cmap='turbo', vmin=0, vmax=1)
        fig.colorbar(axim, label='Strength  [-]')
        axs.xaxis.tick_top()
        axs.xaxis.set_label_position('top')
        axs.set_xlabel('Pixels')
        axs.set_ylabel('Pixels')
        axs.set_title('Splinter orientation towards impact point')
        # create a contour plot of the orientations, that is overlayed onto the original image
        if config.debug:
            fig.show()
            fig.waitforbuttonpress()

        fig.tight_layout()
        fig.savefig(self.__get_out_file(f"fig_splinter_orientation.{config.ext_plots}"))


    def __create_voronoi(self, config: AnalyzerConfig):
        centroids = [x.centroid_px for x in self.splinters if x.has_centroid]
        voronoi = Voronoi(centroids)
        fig = voronoi_plot_2d(voronoi, show_points=True, point_size=5, show_vertices=False, line_colors='red')

        plt.imshow(self.original_image)
        plt.axis('off')  # Turn off axis labels and ticks
        plt.title('Voronoi Plot Overlay on Image')
        if config.debug_experimental:
            plt.show()
        fig.savefig(self.__get_out_file(f"voronoi.{config.ext_plots}"))

        events = np.array(centroids)
        region = np.array([[0,self.original_image.shape[1]], [0, self.original_image.shape[0]]])

        # optimal_h = estimate_optimal_h(events, region)
        # print(f'Optimal h: {optimal_h}')

        #TODO: compare voronoi splinter size distribution to actual distribution
        # X,Y,Z = csintkern(events, region, 500)
        X, Y, Z = csintkern_optimized(events, region, config.intensity_h)
        fig,axs = plt.subplots()
        axs.imshow(self.image_contours)
        axim = axs.contourf(X, Y, Z, cmap='turbo', alpha=0.5)
        fig.colorbar(axim, label='Intensity [Splinters / Area]')
        # plt.scatter([x[0] for x in centroids], [x[1] for x in centroids], color='red', alpha=0.5, label='Data Points')
        # plt.legend()
        axs.xaxis.tick_top()
        axs.xaxis.set_label_position('top')
        axs.set_xlabel('Pixels')
        axs.set_ylabel('Pixels')
        axs.set_title(f'Fracture Intensity (h={config.intensity_h:.2f})')

        if config.debug_experimental:
            plt.show()

        fig.tight_layout()
        fig.savefig(self.__get_out_file(f"fig_intensity.{config.ext_plots}"))

    def __filter_dark_spots(self, config: AnalyzerConfig):
        # create normal threshold of original image to get dark spots
        img = preprocess_spot_detect(self.original_image)
        if config.debug_experimental:
            cimg = cv2.cvtColor(self.original_image, cv2.COLOR_GRAY2BGR)
        i_del = []
        removed_splinters: list[Splinter] = []
        for i,s in enumerate(bar := tqdm(self.splinters, leave=False)):
            x, y, w, h = cv2.boundingRect(s.contour)
            roi_orig = img[y:y+h, x:x+w]

            mask = np.zeros_like(img)
            cv2.drawContours(mask, [s.contour], -1, 255, thickness=cv2.FILLED)

            roi = mask[y:y+h, x:x+w]
            # Apply the mask to the original image
            result = cv2.bitwise_and(roi_orig, roi_orig, mask=roi)

            # Check if all pixels in the contour area are black
            if np.mean(result) < 0.25:
                bar.set_description(f'Removed {len(i_del)} Splinters')
                i_del.append(i)
            elif config.debug_experimental:
                cv2.drawContours(cimg, [s.contour], -1, rand_col(), 1)

        # remove splinters starting from the back
        for i in sorted(i_del, reverse=True):
            removed_splinters.append(self.splinters[i])
            del self.splinters[i]

        if config.debug:
            print(f"Removed {len(removed_splinters)} Splinters")

        skel_mask = self.image_skeleton.copy()

        for s in tqdm(removed_splinters, leave=False):
            c = s.centroid_px

            # Remove the original contour from the mask
            cv2.drawContours(skel_mask, [s.contour], -1, 0, 1)
            cv2.drawContours(skel_mask, [s.contour], -1, 0, -1)

            # cv2.drawContours(skel_mask, [s.contour], -1, 0, -1)
            connections = []

            # Search for adjacent lines that were previously attached
            #   to the removed contour
            for p in s.contour:
                p = p[0]
                # Search the perimeter of the original pixel
                for i,j in [(-1,-1), (-1,0), (-1,1),\
                            (0,-1), (0,1),\
                            (1,-1), (1,0), (1,1) ]:
                    x = p[0] + j
                    y = p[1] + i
                    if x >= self.image_skeleton.shape[1] or x < 0:
                        continue
                    if y >= self.image_skeleton.shape[0] or y < 0:
                        continue

                    # Check if the pixel in the skeleton image is white
                    if skel_mask[y][x] != 0:
                        # Draw a line from the point to the centroid
                        connections.append((x,y))

                        if config.debug_experimental:
                            cv2.drawMarker(cimg, (x,y), (0,255,0))

            # 2 connections -> connect them
            if len(connections) == 2:
                cv2.drawContours(self.image_skeleton, [s.contour], -1, (0), -1)
                x,y = connections[0]
                a,b = connections[1]

                cv2.line(self.image_skeleton, (int(x), int(y)), (int(a), int(b)), 255)

            # more than 2 -> connect each of them to centroid
            elif len(connections) > 2:
                cv2.drawContours(self.image_skeleton, [s.contour], -1, (0), -1)
                # cv2.drawContours(self.image_skeleton, [s.contour], -1, (0), 1)
                for x,y in connections:
                    cv2.line(self.image_skeleton, (int(x), int(y)), (int(c[0]), int(c[1])), 255)

                    if config.debug_experimental:
                        cv2.line(cimg, (int(x), int(y)), (int(c[0]), int(c[1])), (255,0,0))


        if config.debug_experimental:
            cv2.imwrite(self.__get_out_file("spots_filled.png"), cimg)
            plt.imshow(cimg)
            plt.show()


    def __get_out_file(self, file_name: str, file_ext: str = None) -> str:
        """Returns an absolute file path to the output directory.

        Args:
            file_name (str): The filename inside of the output directory.
            file_ext (str): The file extension.
        """
        if file_ext is not None:
            file_name = f'{file_name}.{file_ext}'

        path = os.path.join(self.out_dir, file_name)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        return path

    def __onselect(self,eclick, erelease):
        """ Private function, internal use only. """
        x1, y1 = eclick.xdata, eclick.ydata
        x2, y2 = erelease.xdata, erelease.ydata
        selected_region = (x1, y1, x2, y2)

        if x1 > x2:
            x1,x2 = x2,x1
        if y1 > y2:
            y1,y2 = y2,y1

        self.ax1.set_xlim(x1, x2)
        self.ax1.set_ylim(y1, y2)
        self.ax2.set_xlim(x1, x2)
        self.ax2.set_ylim(y1, y2)
        self.ax3.set_xlim(x1, x2)
        self.ax3.set_ylim(y1, y2)
        plt.draw()
        # print(selected_region)
        return selected_region

    def __create_splintersize_filled_image(self, config: AnalyzerConfig):
        img = self.original_image.copy()
        areas = [x.area for x in self.splinters]

        min_area = np.min(areas)
        max_area = np.max(areas)

        for s in self.splinters:
            clr = get_color(s.area, min_value=min_area, max_value=max_area, colormap_name='turbo')
            cv2.drawContours(img, [s.contour], -1, clr, -1)

        cv2.imwrite(self.__get_out_file("img_splintersizes", config.ext_imgs), img)
        combined = cv2.addWeighted(255-self.original_image, 1.0, img, 0.75, 0.0)
        cv2.imwrite(self.__get_out_file(f"img_splintersizes_combined.{self.config.ext_imgs}"), combined)
        combined = cv2.addWeighted(self.image_skeleton_rgb, 1.0, to_rgb(img), 0.6, 0.0)
        cv2.imwrite(self.__get_out_file(f"debug_skeleton_sizes_combined.{self.config.ext_imgs}"), combined)


    def __plot_backend(self, region = None, display = False) -> None:
        """
        Plots the analyzer backend.
        Displays the original img, preprocessed img, and an overlay of the found cracks
        side by side in a synchronized plot.
        """
        self.fig_comparison, (self.ax3, self.ax1, self.ax2) = \
            plt.subplots(1, 3, figsize=(12, 6))

        # Display the result from Canny edge detection
        self.ax3.imshow(self.original_image)
        self.ax3.set_title("Original")
        self.ax3.axis('off')

        # Display the result from Canny edge detection
        self.ax1.imshow(self.preprocessed_image, cmap='gray')
        self.ax1.set_title("Preprocessed image")
        self.ax1.axis('off')

        # Overlay found contours on the original image
        self.ax2.set_title("Detected Cracks")
        self.ax2.axis('off')
        self.ax2.imshow(self.image_contours)

        if region is not None:
            (x1, y2, x2, y1) = region
            self.ax1.set_xlim(x1, x2)
            self.ax1.set_ylim(y1, y2)
            self.ax2.set_xlim(x1, x2)
            self.ax2.set_ylim(y1, y2)
            self.ax3.set_xlim(x1, x2)
            self.ax3.set_ylim(y1, y2)

        # Connect the rectangle selector to synchronize zooming
        rs = RectangleSelector(self.ax1, self.__onselect, useblit=True, \
            button=[1], spancoords='pixels', )
        rs1 = RectangleSelector(self.ax2, self.__onselect, useblit=True, \
            button=[1], spancoords='pixels', )
        rs2 = RectangleSelector(self.ax3, self.__onselect, useblit=True, \
            button=[1], spancoords='pixels', )
        rs.add_state('square')

        rs1.add_state('square')
        rs2.add_state('square')
        plt.tight_layout()

        if display:
            plt.show()

        self.fig_comparison.savefig(self.__get_out_file(f"fig_comparison.{self.config.ext_plots}"))
        return self.fig_comparison

    def __plot_splintersize_accumulation(self, display = False) -> Figure:
        """Plots a graph of accumulated share of area for splinter sizes.

        Returns:
            Figure: The figure, that is displayed.
        """
        areas = [x.area for x in self.splinters]
        total_area = np.sum(areas)

        # ascending sort, smallest to largest
        areas.sort()

        data = []

        area_i = 0
        for area in areas:
            area_i += area

            p = area_i / total_area
            data.append((area, p))

        data_x, data_y = zip(*data)

        self.fig_area_sum, ax = plt.subplots()

        ax.plot(data_x, data_y, 'g-')
        ax.plot(data_x, data_y, 'rx', markersize=2)
        ax.set_title('Splinter Size Accumulation')
        ax.set_xlabel(f"$Area_i [{'mm' if self.config.cropped_image_size is not None else 'px'}²]$")
        ax.set_ylabel(r"$\frac{\sum (Area_i)}{Area_t} [-]$")
        # plt.axvline(np.average(areas),c='b', label = "Area avg")
        # plt.axvline(np.sum(areas), c='g', label='Found Area')
        # plt.axvline(5000**2, c='r', label='Image Area')
        if display:
            plt.show()

        self.fig_area_sum.savefig(self.__get_out_file(f"fig_sumofarea.{self.config.ext_plots}"))

        return self.fig_area_sum

    def __plot_splintersize_distribution(self, display = False) -> Figure:
        """Plots a graph of Splinter Size Distribution.

        Returns:
            Figure: The figure, that is displayed.
        """
        areas = [x.area for x in self.splinters]

        # ascending sort, smallest to largest
        areas.sort()

        data = []
        for area in np.linspace(np.min(areas),np.max(areas),50):
            index = next((i for i, value in enumerate(areas) if value > area), None)
            p = index
            data.append((area, p))

        data_x, data_y = zip(*data)

        self.fig_area_distr, ax = plt.subplots()
        ax.plot(data_x, data_y, 'g-')
        ax.plot(data_x, data_y, 'ro', markersize=3)
        ax.set_title('Splinter Size Distribution')
        ax.set_xlabel(f"Splinter Area [{'mm' if self.config.cropped_image_size is not None else 'px'}²]")
        ax.set_ylabel(r"Amount of Splinters [-]")
        # plt.axvline(np.average(areas),c='b', label = "Area avg")
        # plt.axvline(np.sum(areas), c='g', label='Found Area')
        # plt.axvline(5000**2, c='r', label='Image Area')
        if display:
            plt.show()

        self.fig_area_distr.savefig(self.__get_out_file(f"fig_distribution.{self.config.ext_plots}"))
        return self.fig_area_distr



from __future__ import annotations

import argparse
import base64
import glob
import os
import xml.etree.ElementTree as ET
from collections import defaultdict
from typing import Tuple, TypeVar

import numpy as np

from fracsuite.scalper.scalp_stress import calculate_simple

descr = """
███████╗ ██████╗ █████╗ ██╗     ██████╗ ███████╗██████╗ 
██╔════╝██╔════╝██╔══██╗██║     ██╔══██╗██╔════╝██╔══██╗
███████╗██║     ███████║██║     ██████╔╝█████╗  ██████╔╝
╚════██║██║     ██╔══██║██║     ██╔═══╝ ██╔══╝  ██╔══██╗
███████║╚██████╗██║  ██║███████╗██║     ███████╗██║  ██║
╚══════╝ ╚═════╝╚═╝  ╚═╝╚══════╝╚═╝     ╚══════╝╚═╝  ╚═╝
Leon Bohmann     TUD - ISMD - GCC        www.tu-darmstadt.de/glass-cc
                                                      
                                                        
ScalpProject:           Create with project file. This will split all measurements into 
                        specimens and then into locations.
Specimen:               All measurements on a single specimen.
MeasurementLocation:    One point on the ply that has been measured (3+ Measurements)
Measurement:            A measurement at a location in one specific direction.


This script iterates either one single project or recursively searches for projects in a folder.
For every .scp File, a ScapProject is created, which will handle the reading of data
and also the grouping into Specimens and MeasurementLocations and Measurement respectively.

At the end of the script, all specimens are extracted for better organisation, so that
individual specimen tasks can be performed. For example the calculation of prinicipal stresses
and the mean value of all measurements.

"""

SCALP_ANGLE = 71.5                          # scalp laser angle
SCALP_C = 2.72e-6                           # fracturing index
SWITCH_REMOVE_INVALID_LOCATIONS = False     # delete invalid measurement locations
                                            #   (invalid measurements are not used for stress calculation)
SWITCH_DISPLAY_MOHR_CIRCLE = False
class ScalpProject:
    """
    Main object that is used to wrap a scalp project file (.scp).
    
    Contains a list of all measurements and a list of specimens.
    """
    
    filepath: str
    measurements: list[Measurement]
    specimens: list[Specimen]
    
    def __init__(self, path: os.PathLike):
        
        # print(f"Creating project from file: {path}")
        self.filepath: str = path
        self.measurements: list[Measurement] = []
         
        # read the scalp xml file               
        with open(path, "r") as file:
            tree = ET.parse(file)
            root = tree.getroot()
            # each measurement has to be parsed into a measurement object
            for measurement_node in root.findall(".//Measurements/Measurement"):
                self.measurements.append(Measurement.from_node(measurement_node))
                
        
        # group measurements to specimens
        specimen_measurements: dict[str, list[Measurement]] = defaultdict(list)
        for ms in self.measurements:
            specimen_measurements[ms.specimen].append(ms) 
        
        # create a list of all specimens for easy access
        self.specimens = [Specimen(name, measurements) for name, measurements in specimen_measurements.items()]
        
        
    def write_measurements(self, output_directory: str | None, output_extension):
        """Writes the measurements to an output directory.

        Args:
            output_directory (str): Output path
            output_extension (str): Output extension
        """
        for measurement in self.measurements:
            output_path = os.path.join(output_directory, measurement.specimen, 'scalp')
            os.makedirs(output_path, exist_ok=True)
            output_path = os.path.join(output_path, f"{measurement.name}.{output_extension}")
            with open(output_path, "w") as file:
                measurement.to_file(file)

        for specimen in self.specimens:
            output_path = os.path.join(output_directory, specimen.name, 'scalp')
            output_path = os.path.join(output_path, f"{specimen.name}_stress.txt")
            with open(output_path, "w") as file:
                specimen.to_file(file)
            
        
class Specimen:
    """ Object, that contains information about a specimen. """
    name: str
    measurementlocations: list[MeasurementLocation]
    
    measured_thickness: float   # measured thickness
    sig_h: float                # homogenous pre-stress value
    sig_h_dev: float            # standard deviation of pre-stress
    
    nue: float = 0.23
    E: float = 72.0e3
    
    def __init__(self, name: str, measurements: dict[str,list[Measurement]]):
        self.name = name
        self.sig_h = 0.0
                        
        # print(f"Extracting info for specimen {name}")
        
        # group measurements to locations
        locations_measurements: dict[str, list[Measurement]] = defaultdict(list)
        for ms in measurements:
            locations_measurements[ms.location].append(ms)
           
        self.measurementlocations = \
            [MeasurementLocation(loc, name, x) \
                for loc, x in locations_measurements.items()]
        
        # invalid locations can be removed, if the constant is set to TRUE
        #   even if the locations are not removed, they will not be used for calculation
        if SWITCH_REMOVE_INVALID_LOCATIONS:
            # remove invalids
            for loc in self.measurementlocations:
                if loc.invalid:
                    self.measurementlocations.remove(loc)
        
        
        # get homogenous stress level of the specimen
        self.__calc_homogenous_princ_stress()
        
        # get thickness data from measurements
        thickness = 0.0
        for meas in measurements:
            if thickness != 0.0 and thickness != meas.measured_thickness:
                print(f"WARNING: Different thickness measurements in specimen {name}.")
            
            thickness = meas.measured_thickness
           
        self.measured_thickness = thickness 
        
        # calculate strain energy density
        self.U_d = 1/5 * (1-self.nue)/self.E * self.sig_h ** 2
    
    def __calc_homogenous_princ_stress(self):
        """
        Calculates the homogenous principal stress by taking the 
        mean value of all measured locations. Invalid locations (those
        with missing directions like ur, m or ol) will not contribute
        to the mean value.
        """
        stresses1 = []
        stresses2 = []
        for location in self.measurementlocations: 
            # dont use invalid locations for stress measurement
            if location.invalid:
                continue
                   
            stresses1.append(location.stress[0])
            stresses2.append(location.stress[1])
        
        self.sig_h = np.mean(stresses1 + stresses2) 
        self.sig_h_dev = np.std(stresses1 + stresses2)       
    
    def to_file(self, file):
        file.write(f'{self.measured_thickness:<35}\t# thickness [mm]\n')   
        file.write(f'{self.sig_h:<35}\t# homogenous stress [MPa]\n')   
        file.write(f'{self.sig_h_dev:<35}\t# hom stress std-dev [-]\n')
        file.write(f'{self.U_d:<35}\t# energy density [MPa]\n')
        file.write('\n')
        
        file.write(f'# {"name":18}\t{"sig_1":<20}\t{"sig_2":<20}\n')
        
        for loc in self.measurementlocations:
            file.write(f'{loc.location_name:20}\t{loc.stress[0]:<20}\t{loc.stress[1]:<20}\n')
    
    def __str__(self):
        return self.name

class MeasurementLocation:
    """
    This class bundles all measurements (in different directions) at a location
    on the glass ply.
    Typically, this class contains the X, Y and XY direction. If any of the directions
    is missing, the location is marked as "invalid" and the measurement will not
    contribute to the calculation of the stresses on the ply.
    """
    location_name: str
    specimen_name: str
    measurements: list[Measurement]
    directions_stress: dict[str, Measurement]
    stress: Tuple[float,float]
    invalid: bool
        
    def __init__(self, location_name: str, specimen_name: str, measurements: list[Measurement]):
        self.location_name = location_name
        self.specimen_name = specimen_name
        self.measurements = measurements
        
        # save measurements in dictionary      
        self.directions_stress = {}  
        self.directions_stress["X"] = None
        self.directions_stress["Xy"] = None
        self.directions_stress["Y"] = None
        
        for meas in self.measurements:
            self.directions_stress[meas.orientation] = meas.stress
        
        
        # All 3 directions must be present in order to calculate stresses
        #   For every missing direction, stress state can be entered manually.
        #   If the stress state is not known, the stress of this location can
        #   can not contribute to the stress calculation of the specimen.
        if not self.__has_direction("X") \
            or not self.__has_direction("XY") \
                or not self.__has_direction("Y"):
            print(f"> WARNING: {specimen_name}-{location_name} is missing one or more directions!")
            
            for missing_dir in self.__get_missing_directions():
                sig = input(f"Enter stress for {specimen_name}-{location_name}, Direction {missing_dir} [sigma]: ")   
                
                # if no stress is applied, the location is invalid
                if sig == '':
                    self.invalid = True
                    self.stress = (np.nan, np.nan)
                    return
                             
                sig = float(sig)
                self.directions_stress[missing_dir] = [sig]
                self.invalid = False
        else:
            self.invalid = False        
            
        # this is only reached, if all directions are present as measurements
        # or if missing measurements are replaced by manual measurement inputs
        self.stress = calculate_simple(self.directions_stress['X'], \
            self.directions_stress['Y'], self.directions_stress['Xy'], \
                np.deg2rad(SCALP_ANGLE), \
                    SCALP_C, SWITCH_DISPLAY_MOHR_CIRCLE)

    def __has_direction(self, dir: str) -> bool:
        for meas in self.measurements:
            if meas.orientation.lower() == dir.lower():
                return True
        
        return False

    def __get_direction(self, dir: str) -> Measurement | None:
        for meas in self.measurements:
            if meas.orientation.lower() == dir.lower():
                return meas
        
        return None

    def __get_missing_directions(self) -> list[str]:
        dirs = ["X", "Xy", "Y"]
        missing = []
        
        for dir in dirs:
            if not self.__has_direction(dir):
                missing.append(dir)
        
        return missing
        
class Measurement:
    """
    Contains all raw data from a SCALP Measurement.
    """
    
    ## Attributes
    name: str
    specimen: str                   # z.B. 4.70.Z.1
    location: str                   # ul, m, or
    orientation: str                # x, y, xy    
    measured_thickness: float
    
    ## Data
    data: list[tuple[str,float,float]] # Stress Value Name, Value, Sigma
    depth: list[float]
    stress: list[float]
    active: list[bool]
    fit: list[bool]    
    retardation: list[list[float]] # list of all measured retardations in a Measurement
    
    def __init__(self, specimen: str,  loc: str, orientation: str, spec_thickness:float, data: list[tuple[str,float,float]], \
        depth: list[float], stress: list[float], active: list[float], fit: list[float], \
            ret: list[list[float]]):

        self.name = f"{specimen}_{loc}-{orientation}"   
        self.specimen = specimen
        self.orientation = orientation
        self.measured_thickness = spec_thickness
        
        self.data = data
        self.depth = depth
        self.stress = stress
        self.active = active
        self.fit = fit
        self.retardation = ret
        self.location = loc
        
    def from_node(measurement_node) -> Measurement:
        """Creates a Measurement object from a xml Node.

        Args:
            measurement_node (Node): XML Node retrieved from node.find("x").

        Returns:
            Measurement: A measurement object.
        """
        specimen = measurement_node.find("Specimen")
        if specimen is None:
            return None

        specimen_name = specimen.attrib.get("Name")
        specimen_note = specimen.attrib.get("Note")
        specimen_dir = specimen.attrib.get("Direction").replace("-axis", "")
        specimen_thickness = float(specimen.attrib.get("GlassThickness"))
        
        
        stress_distribution = measurement_node.find("StressDistribution")
        if stress_distribution is None:
            return None

        curve_depth = np.frombuffer(base64.b64decode(stress_distribution.find("Depth").text), dtype=np.float32)
        curve_stress = np.frombuffer(base64.b64decode(stress_distribution.find("Stress").text), dtype=np.float32)
        curve_active = np.frombuffer(base64.b64decode(stress_distribution.find("Active").text), dtype=np.bool_)
        curve_fit = np.frombuffer(base64.b64decode(stress_distribution.find("Fit").text), dtype=np.int8)
        
        # retardation is not fitted and averaged over the measured data
        #   So, every measured retardation is read and saved as a list of curves, so that
        #   we can later decide how to use it
        retardation_data = measurement_node.findall(".//MeasuredData/Data/StressDistribution/Retardation")
        # print(retardation_data)
        retardation_data_list: list[list[float]] = []
        
        for data in retardation_data:
            
            # retardation_node = data.find("StressDistribution").find("Retardation")
            # if retardation_node is None:
            #     print(data)
            #     continue
            
            retardation_text = data.text
            
            curve_retardation = np.frombuffer(\
                base64.b64decode(retardation_text),\
                    dtype=np.float32)
            retardation_data_list.append(curve_retardation)
        
        stress_data = []

        stress_values = measurement_node.find("StressValues")
        if stress_values is not None:
            for stress_value in stress_values.iter():
                sv_name = stress_value.tag
                avg = float(stress_value.attrib.get("Col1", 0))
                stddev = float(stress_value.attrib.get("Col2", 0))
                stress_data.append((sv_name, avg, stddev))

        return Measurement(specimen_name,  specimen_note, specimen_dir, specimen_thickness, stress_data, curve_depth, curve_stress, \
            curve_active, curve_fit, retardation_data_list)

    def to_file(self, file):
        file.write(f"# Measurement: {self.name}\n")
        for data in self.data:
            file.write(f"# {data[0]:<20} = {data[1]:<10.4f} (+/- {data[2]:<10.4f})\n")
        file.write("# Depth                Stress               Active               Fit\n")
        for i in range(len(self.depth)):
            file.write(f"{self.depth[i]:<20.4f}{self.stress[i]:<20.4f}{self.active[i]:<20}{self.fit[i]:<20}\n")    
    
T = TypeVar("T")
def flatten_and_sort(l: list[list[T]], sort_mtd) -> list[T]:
    ret = [item for sublist in l for item in sublist]
    ret.sort(key = sort_mtd)
    return ret

def remove_duplicates(my_list, my_lambda):
    new_list = []
    for item in my_list:
        exists = False
        for ex_item in new_list:
            if my_lambda(ex_item) == my_lambda(item):
                exists = True
                print(f'Found and removed duplicate specimen: {item}')

        if exists:
            continue
                
        new_list.append(item)
        
    return new_list


def get_specimens_from_projects(projects) -> list[Specimen]:
    """Extract all specimens in a distinct list from a list of projects.

    Args:
        projects (list[ScalpProject]): List of scalp projects.

    Returns:
        list[Specimen]: Ordered and distinct list of all specimens.
    """
    specimens = flatten_and_sort([x.specimens for x in projects], lambda s: s.name)
    specimens = remove_duplicates(specimens, lambda t: t.name)
    
    return specimens

########################################################################################
#   SCRIPT AREA
#
#


# implement parse to make this script callable from outside
parser = argparse.ArgumentParser(description=descr, formatter_class=argparse.RawTextHelpFormatter)
parser.add_argument('-project', nargs="?", help='The project to be processed.', default=None)  
parser.add_argument('-output', nargs="?", help='Output directory.')  
parser.add_argument('-folder', nargs="?", help='If this is passed, analyzes all scalp projects in a folder.', default=None)  
parser.add_argument('--mohr', action='store_true', help='Display mohr"sche circle for stress states.')
args = parser.parse_args()

# overwrite global vars
SWITCH_DISPLAY_MOHR_CIRCLE = args.mohr

# input vars
project_file = args.project
output_directory = args.output

output_extension = "txt"
open_output_directory = True

# default project file
if project_file is None and args.folder is None:
    raise Exception("No project file or folder supplied!")
# default project file
if output_directory is None:
    if project_file is not None:
        output_directory = os.path.join(os.path.dirname(project_file), "out")
    elif args.folder is not None:
        output_directory = os.path.join(os.path.dirname(args.folder), "out")

# create the output folder if it does not exist
if not os.path.exists(output_directory):
    os.makedirs(output_directory)

projects: list[ScalpProject] = []

# analyze all projects in the folder
if args.folder is not None:
    project_folder = args.folder
    extension = '.scp'
    files = glob.glob(os.path.join(project_folder, '**', f'*{extension}'), recursive=True)
    
    for file in files:
        print(f'Analyze file: {file.replace(project_folder, "")}')
        project = ScalpProject(file)
        project.write_measurements(output_directory, output_extension)
        projects.append(project)
# analyze one project
else:
    project = ScalpProject(project_file)
    project.write_measurements(output_directory, output_extension)    

    projects.append(project)
    
    
# write a summary    
project_len = len(projects)
measurements_len = np.sum([len(x.measurements) for x in projects])

print(f'Projects       : {project_len}')
print(f'Measurements   : {measurements_len}')

# output all specimens independently from their project    
specimens = get_specimens_from_projects(projects)
for specimen in specimens:
    
    pre = ""
    if not all(not t.invalid for t in specimen.measurementlocations):
        pre = "!!!"
        
    print(f'\t{pre}\t{specimen.name:10}: {len(specimen.measurementlocations)} Locations ({[(x.location_name, [l.orientation for l in x.measurements]) for x in specimen.measurementlocations]})')
    
    
# calculate stresses for specimens
for specimen in specimens:

    print(f'\t{specimen.name:10}: sig_h={specimen.sig_h:.2f} (+- {specimen.sig_h_dev:.2f})MPa')
    print('\t\t', end="")
    
    for loc in specimen.measurementlocations:
        print(f'\t{loc.location_name}={loc.stress[0]:.2f}/{loc.stress[1]:.2f}', end = "")
        
    print()


import os
import cv2
import re
import shutil

from argparse import ArgumentParser
from itertools import groupby
from fracsuite.identifier.barcode import dispImage

from fracsuite.splinters.analyzer import crop_perspective
from fracsuite.identifier import read_barcode
from fracsuite.splinters.analyzerConfig import AnalyzerConfig

from rich import print

def get_unique_scan_id(file) -> str:
    """Get the first 4 numbers of cullet scanner file name.
    Format is: [ID] [Date] ([img_type]).[ext]

    Args:
        file (str): Full file path.

    Returns:
        str: File identifier.
    """
    return os.path.basename(file)[:4]

def get_scan_type(input: str) -> str:
    """
    Function to extract content between the last two brackets in a filename.
    Args:
        filename (str): The filename from which content is to be extracted.
    Returns:
        str: The content between the last two brackets.
    """
    # Find all substrings that match the pattern
    matches = re.findall(r'\[.*?\]', input)

    # If there are matches, return the content of the last match
    # without the brackets. Otherwise, return an empty string.
    if matches:
        return matches[-1][1:-1]
    else:
        return ''

parser = ArgumentParser()
parser.add_argument('directory', type=str, help='Directory with scans.')
parser.add_argument('--dry', action="store_true", help='Perform dry run that only reads codes.')
parser.add_argument('-debug', choices=['label', 'code', 'barcode', 'error'], nargs="*", help='Show debug images.')

args = parser.parse_args()

if args.debug:
    print("[yellow]Debug mode enabled![/yellow]")

config = AnalyzerConfig()


root_dir = args.directory


# read input folder and group files
extensions = [".bmp", ".zip"]   
files: list[str] = [os.path.join(root_dir, f) \
        for f in os.listdir(root_dir) \
            if os.path.isfile(os.path.join(root_dir, f)) \
            and any(f.endswith(ext) for ext in extensions)]

sorted_files = sorted(files, key=get_unique_scan_id)
# group files for file id
grouped_files = {k: list(g) for k, g in groupby(sorted_files, key=get_unique_scan_id)}

# create archive folder
archive_folder = os.path.join(root_dir, ".archive")
os.makedirs(archive_folder, exist_ok=True)

for key, group in grouped_files.items():
    if not any(file.endswith(".bmp") and "Transmission" in file for file in group):
        print(f"Couldn't find Transmission-Scan for scan ID {key}!")
        continue
    
    # get bitmaps from group
    img0_path = next(file for file in group if file.endswith(".bmp") and "Transmission" in file)
    # load image and perform OCR to find specimen Identifier ([thickness].[residual_stress].[boundary].[ID])
    img0 = cv2.imread(img0_path)
    img0 = crop_perspective(img0, None, False) 
    img0 = cv2.rotate(img0, cv2.ROTATE_90_COUNTERCLOCKWISE)
    dispImage(img0, "Original", "Main", args.debug)
    
    series = read_barcode(img0, args.debug)
    
    # if None, Datamatrix could not be read
    if series is None:
        print(f"Couldn't find series for scan ID {key}!")
        continue
        
    print(f"Found {series} for ID {key}...")
    
    if args.dry:
        continue
    
    series = os.path.join(root_dir, series)
    subfolder = os.path.join(series, "anisotropy")    
    if os.path.exists(subfolder):
        print(f"Scan ID {key} already processed!")
        continue
    
    # create a subfolder for the scans called "anisotropy"
    os.makedirs(subfolder, exist_ok=True)
    
    # iterate over each file in group, run perspective transform on bitmaps only and copy all into the subfolder
    for file in group:
        if file.endswith(".bmp"):
            img = cv2.imread(file)
            img = crop_perspective(img, (4000,4000), False)
            img = cv2.rotate(img, cv2.ROTATE_90_COUNTERCLOCKWISE)
            cv2.imwrite(os.path.join(subfolder, f'{get_scan_type(file)}.bmp'), img)
            
            # move source image to archive
            os.rename(file, os.path.join(archive_folder, os.path.basename(file)))
        else:
            # copy file to archive
            shutil.copy(file, os.path.join(archive_folder, os.path.basename(file)))
            
            os.rename(file, os.path.join(subfolder, os.path.basename(file)))



"""
███████╗██████╗ ██╗     ██╗███╗   ██╗████████╗███████╗██████╗ ███████╗
██╔════╝██╔══██╗██║     ██║████╗  ██║╚══██╔══╝██╔════╝██╔══██╗██╔════╝
███████╗██████╔╝██║     ██║██╔██╗ ██║   ██║   █████╗  ██████╔╝███████╗
╚════██║██╔═══╝ ██║     ██║██║╚██╗██║   ██║   ██╔══╝  ██╔══██╗╚════██║
███████║██║     ███████╗██║██║ ╚████║   ██║   ███████╗██║  ██║███████║
╚══════╝╚═╝     ╚══════╝╚═╝╚═╝  ╚═══╝   ╚═╝   ╚══════╝╚═╝  ╚═╝╚══════╝
                                                                      
Leon Bohmann     TUD - ISMD - GCC        www.tu-darmstadt.de/glass-cc

Description:
-------------------------
This module helps with the automated detection of fragments on fractured glass plys. It
performs some preprocessing actions to improve the quality of the input image and then
uses that to search for contours on the image. After filtering and further improvement
the found contours are converted into `Splinter` objects, which allow further investigation
of the fragments (size, roughness, roundness, ...).

Used packages:
-------------------------
- opencv-python
- matplotlib
- numpy
- tqdm    

Usage:
-------------------------                       
Command line usage is shown below. For further information visit:
https://github.com/leonbohmann/fracture-suite
"""

import os
from fracsuite.splinters.analyzer import Analyzer
from fracsuite.splinters.analyzerConfig import AnalyzerConfig

parser = AnalyzerConfig.get_parser(__doc__)

parser.add_argument("--all", default=False, 
                    help="Instruct the analyzer to run the analysis on every subfolder.",
                    action='store_true')

args = parser.parse_args()
config = AnalyzerConfig.from_args(args)

if args.all:
    project_dir = config.path
    for file in os.listdir(project_dir):
        project_path = os.path.join(project_dir, file, '\\')
        
        if os.path.exists(project_path):
            config.path = project_path
            analyzer = Analyzer(config)
        
        
        
else:
    analyzer = Analyzer(config)
    




import numpy as np
import cv2


class Splinter:

    angle: float
    "Orientation of the splinter in degrees."
    angle_vector: tuple[float, float]
    "Normalized orientation vector"
    alignment_score: float = np.nan
    "Score indicating how much the splinter points into the direction of the impact point."

    def __init__(self, contour, index, mm_px: float, config = None):
        """Create a splinter from a contour.

        Args:
            contour (np.array): Input contour.
            index (int): Index of the splinter.
            mm_px (float): Scale factor for area. px/mm.
        """
        self.ID = index
        self.contour = contour

        self.area = cv2.contourArea(self.contour) * mm_px ** 2
        self.circumfence = cv2.arcLength(self.contour, True) * mm_px

        # roundness
        self.roundness = 4 * np.pi * self.area / self.circumfence ** 2
        # roughness
        self.roughness = self.calculate_roughness()

        # centroid
        try:
            M = cv2.moments(self.contour)
            cX = int(M["m10"] / M["m00"])
            cY = int(M["m01"] / M["m00"])

            self.centroid_mm =  (cX * mm_px, cY * mm_px)
            self.centroid_px =  (cX, cY)
        except:
            self.centroid_mm = (np.nan, np.nan)
            self.centroid_px = (np.nan, np.nan)


        self.has_centroid = not any(np.isnan(self.centroid_mm)) and not any(np.isnan(self.centroid_px))


        self.angle = self.__calculate_orientation()
        # try:
        #     _,_, self.angle = cv2.fitEllipse(self.contour)
        # except:
        #     self.angle = np.nan

        if config is not None:
            self.measure_orientation(config)

    def calculate_roughness(self) -> float:
        """Calculate the roughness of the contour by comparing the circumfence
        of the contour to the circumfence of its convex hull.

        Returns:
            float: A value indicating how rough the perimeter is.
        """
        contour = self.contour
        perimeter = cv2.arcLength(contour,True)
        hull = cv2.convexHull(contour)
        hullperimeter = cv2.arcLength(hull,True)

        return perimeter / hullperimeter

    def __calculate_orientation(self):
        """Calculate the orientation of the splinter in degrees."""
        M = cv2.moments(self.contour)

        # Calculate the angle
        angle = 0.5 * np.arctan2(2 * M['mu11'], M['mu20'] - M['mu02'])
        angle_degrees = np.degrees(angle)

        if angle_degrees < 0:
            angle_degrees += 180

        angle_radians = np.radians(angle_degrees)
        x = np.cos(angle_radians)
        y = np.sin(angle_radians)
        self.angle_vector = np.array((x, y))
        return angle_degrees

    def in_region(self, rect: tuple[float,float,float,float]) -> bool:
        """Check if the splinter is in the given region.

        Args:
            rect (tuple[float,float,float,float]): Region to check. (x1,y1,x2,y2)

        Returns:
            bool: True if the splinter is in the region.
        """
        x1,y1,x2,y2 = rect
        x,y = self.centroid_mm
        return x1 <= x <= x2 and y1 <= y <= y2

    def in_region_exact(self, rect: tuple[float,float,float,float]) -> bool:
        x1, y1, x2, y2 = rect
        for point in self.contour:
            x, y = point[0]
            if x < x1 or x > x2 or y < y1 or y > y2:
                return False
        return True

    def __calculate_orientation_score(self, origin) -> float:
        """Calculate the alignment score of the splinter with the given vector.

        Returns:
            A value from [0,1] indicating, how much the splinter points into the direction of the given vector.
        """
        centroid = self.centroid_mm
        dx = origin[0] - centroid[0]
        dy = origin[1] - centroid[1]

        angle_radians = np.deg2rad(self.angle)
        line_direction = np.array([dx, dy])
        angle_vector = np.array([np.cos(angle_radians), np.sin(angle_radians)])
        dot_product = np.dot(line_direction, angle_vector)
        magnitude_line = np.linalg.norm(line_direction)
        self.alignment_score = np.abs(dot_product) / magnitude_line
        return self.alignment_score



    def measure_orientation(self, config) -> float:
        """Calculate, how much the splinters orientation points to the impactpoint of config.

        Args:
            config (AnalyzerConfig): Configuration object.

        Returns:
            float: Orientation in degrees.
        """
        if not self.has_centroid:
            return np.nan

        # calculate the angle between the centroid and the impact point

        return self.__calculate_orientation_score(config.impact_position)



from typing import Tuple
import numpy as np
from matplotlib import pyplot as plt

descr = """
Calculate actual stress state at the surface of a scalp measured glass ply.

This takes three functions of [Depth, Stress]. They can be obtained from the scalp
project files either by converting them to txt or by using scalper.py to create
text files with the needed values automatically.

Specimen identification has to look as follows: THICKNESS.NOM_STRESS.BOUND.ID
From this format the needed information are read, such as the thickness.
"""

# read the functions of stress over depth
# fit a function into the points
# create a derivation of the functions
# apply magic


if __name__ == "__main__":
    # data: [[depth], [stress]]
    sig_x_data = []
    sig_y_data = []
    sig_xy_data = []
    
    
    
    pass

def calculate_simple(meas_x: list[float], meas_y: list[float], meas_xy: list[float], \
    beta: float, C: float, mohr: False) -> Tuple[float,float]:
    """
    Calculate principal stresses assuming planar and isotropic stress state.
    Uses the stress values obtained from scalp measurements directly.

    Args:
        meas_x (list[float]): Data in x-Direction.
        meas_y (list[float]): Data in y-Direction.
        meas_xy (list[float]): Data in xy-Direction.
        beta (float): Angle of the scalp Laser.
        C (float): Lightfracturing constant of glass ply.
    
    Returns:
        Tuple with (sig_1, sig_2).
    """
    # evaluate at surface
    sig_x_0 = meas_x[0]
    sig_y_0 = meas_y[0]
    sig_xy_0 = meas_xy[0]
    
    # simplified approach for principal stresses
    t_xy = sig_xy_0 - (sig_y_0 + sig_x_0) / 2
    phi1 = 0.5 * np.arctan(2*t_xy/(sig_x_0 - sig_y_0))
    phi2 = phi1 + np.pi / 2
    
    sig_1 = (sig_x_0 + sig_y_0) / 2 + ((sig_x_0 - sig_y_0) / 2) * np.cos(2*phi1) + t_xy * np.sin(2*phi1)
    sig_2 = (sig_x_0 + sig_y_0) / 2 + ((sig_x_0 - sig_y_0) / 2) * np.cos(2*phi2) + t_xy * np.sin(2*phi2)
    
    
    if np.abs(sig_x_0) > np.abs(sig_y_0):
        s1 = sig_y_0
        s2 = sig_x_0
    else:    
        s1 = sig_x_0
        s2 = sig_y_0
    
    # draw_mohrs_stress_circle(s1, s2, sig_x_0, t_xy)
    if mohr:
        draw_mohrs_stress_circle(sig_1, sig_2, s1, np.abs(t_xy), s2, -np.abs(t_xy))

    return sig_1, sig_2

def calculate_simple_ret(d_x: list[float], d_y: list[float], d_xy: list[float], \
    ret_x: list[float], ret_y: list[float], ret_xy: list[float], \
    beta: float, C: float) -> Tuple[float,float]:
    """
    Calculate principal stresses assuming planar and isotropic stress state.
    Uses retardation derivation to calculate stresses.

    Args:
        meas_x (Measurement): Measurement in x-Direction.
        meas_y (Measurement): Measurement in y-Direction.
        meas_xy (Measurement): Measurement in xy-Direction.
        beta (float): Angle of the scalp Laser.
        C (float): Lightfracturing constant of glass ply.
    
    Returns:
        Tuple with (sig_1, sig_2).
    """
    
    # parabolic polyfit on the functions
    ret_x = np.polyfit(d_x, ret_x, 2)
    ret_y = np.polyfit(d_y, ret_y, 2)
    ret_xy = np.polyfit(d_xy, ret_xy, 2)
    
    # derivation of the functions
    dret_x = np.polyder(ret_x)
    dret_y = np.polyder(ret_y)
    dret_xy = np.polyder(ret_xy)
    
    # evaluate at surface
    sig_x_0 = dret_x[0] / ( C * np.cos(beta)**2)
    sig_y_0 = dret_y[0] / ( C * np.cos(beta)**2)
    sig_xy_0 = dret_xy[0] / ( C * np.cos(beta)**2)
    
    # simplified approach for principal stresses
    t_xy = sig_xy_0 - (sig_y_0 - sig_x_0) / 2
    phi1 = 0.5 * np.arctan(2*t_xy/(sig_x_0 - sig_y_0))
    phi2 = phi1 + np.pi / 2
    
    sig_1 = (sig_x_0 + sig_y_0) / 2 + ((sig_x_0 - sig_y_0) / 2) * np.cos(2*phi1) + t_xy * np.sin(2*phi1)
    sig_2 = (sig_x_0 + sig_y_0) / 2 + ((sig_x_0 - sig_y_0) / 2) * np.cos(2*phi2) + t_xy * np.sin(2*phi2)
    
    
    return sig_1, sig_2

def calculate_nlgs(d_x: list[float], d_y: list[float], d_xy: list[float], \
    ret_x: list[float], ret_y: list[float], ret_xy: list[float]):
    """Calculate actual stress using three measurements.

    Args:
        meas_x (Measurement): Measurement in x-Direction.
        meas_y (Measurement): Measurement in y-Direction.
        meas_xy (Measurement): Measurement in xy-Direction.
    """
    
     # parabolic polyfit on the functions
    ret_x = np.polyfit(d_x, ret_x, 2)
    ret_y = np.polyfit(d_y, ret_y, 2)
    ret_xy = np.polyfit(d_xy, ret_xy, 2)
    
    # derivation of the functions
    dret_x = np.polyder(ret_x)
    dret_y = np.polyder(ret_y)
    dret_xy = np.polyder(ret_xy)
    
def draw_mohrs_stress_circle(sigma_1, sigma_2, sigma_n1, tau_n1, sigma_n2, tau_n2):
    # Calculate the center of the circle (average of principal stresses)
    center_stress = (sigma_1 + sigma_2) / 2.0

    # Calculate the radius of the circle
    radius = (sigma_1 - sigma_2) / 2.0

    # Plot the circle
    theta = np.linspace(0, 2*np.pi, 100)
    x_circle = center_stress + radius * np.cos(theta)
    y_circle = radius * np.sin(theta)

    plt.plot(x_circle, y_circle)

    # Plot the principal stresses
    plt.plot([sigma_1, sigma_2], [0, 0], 'ro', label='Principal Stresses')

    # Plot the center stress
    plt.plot(center_stress, 0, 'go', label='Center Stress')

    # Plot the normal stress
    plt.plot(sigma_n1, 0, 'bo', label='σ_a')
    plt.plot(sigma_n2, 0, 'bo', label='σ_b')

    # Connect the center stress with the shear stress on the circle
    plt.plot([center_stress, sigma_n1], [0, tau_n1], 'g--', label='τ_xy')
    plt.plot([center_stress, sigma_n2], [0, tau_n2], 'g--')

    # Connect the normal stress with the shear stress on the circle
    plt.plot([sigma_n1, sigma_n1], [0, tau_n1], 'b--')
    plt.plot([sigma_n2, sigma_n2], [0, tau_n2], 'b--')

    # Set axis limits and labels
    plt.axis('equal')
    plt.xlabel('Normal Stress')
    plt.ylabel('Shear Stress')
    plt.title("Mohr's Stress Circle")

    # Add legend
    plt.legend(bbox_to_anchor=(1.05, 1.0), loc='upper left')
    plt.tight_layout()
    # Show the plot
    plt.grid()
    plt.show()

    
if __name__ == "__main__":
    # parse arguments
    # if input is a file, use file
    # if input is a folder, use on every subfolder and find "scalp" subdirectories
    #   in folders that are named after an ID
    # create a new file in "scalp" folder, that contains the calculations
    
    pass


import argparse
import fracsuite.scalper
from fracsuite.splinters.analyzerConfig import AnalyzerConfig

mainParser = argparse.ArgumentParser(prog="fracsuite", description="A suite of tools for analyzing impact fragmentation images.")



args = mainParser.parse_args()
